# This file is part of pycloudlib. See LICENSE file for license information.
# pylint: disable=too-many-public-methods
"""Base class for all instances to provide consistent set of functions."""

import logging
import time
from abc import ABC, abstractmethod
from contextlib import suppress

import paramiko
from paramiko.ssh_exception import (
    AuthenticationException,
    BadHostKeyException,
    NoValidConnectionsError,
    PasswordRequiredException,
    SSHException,
)

from pycloudlib.errors import PycloudlibTimeoutError
from pycloudlib.result import Result
from pycloudlib.util import shell_pack, shell_quote


class BaseInstance(ABC):
    """Base instance object."""

    _type = "base"

    def __init__(self, key_pair):
        """Set up instance."""
        self._log = logging.getLogger(__name__)
        self._ssh_client = None
        self._sftp_client = None
        self._tmp_count = 0

        self.boot_timeout = 120
        self.key_pair = key_pair
        self.port = "22"
        self.username = "ubuntu"
        self.connect_timeout = 60
        self.banner_timeout = 60

    @property
    @abstractmethod
    def name(self):
        """Return instance name."""
        raise NotImplementedError

    @property
    @abstractmethod
    def ip(self):
        """Return IP address of instance."""
        raise NotImplementedError

    def get_boot_id(self):
        """Get the instance boot_id.

        Returns:
            string with the boot UUID
        """
        result = self.execute(
            "cat /proc/sys/kernel/random/boot_id", no_log=True
        )
        if result.failed:
            raise OSError(
                f"Failed to get boot_id. Return code: {result.return_code}, "
                f"stdout: {result.stdout}, stderr: {result.stderr}"
            )
        return result

    def console_log(self):
        """Return the instance console log.

        Raises NotImplementedError if the cloud does not support fetching the
        console log for this instance.
        """
        raise NotImplementedError

    @abstractmethod
    def delete(self, wait=True):
        """Delete the instance.

        Args:
            wait: wait for instance to be deleted
        """
        raise NotImplementedError

    def restart(self, wait=True, **kwargs):
        """Restart an instance."""
        self._sync_filesystem()
        # If we're not waiting, just call subclass's restart and return.
        if not wait:
            self._do_restart(**kwargs)
            return

        pre_boot_id = None

        # If we attempt to restart, but the instance is already in a
        # non-connectable state, then don't check boot ids.
        try:
            pre_boot_id = self.get_boot_id()
        except (SSHException, OSError):
            # Case 2: wait=True, but the instance is unreachable.
            # The best we can do is to send a reboot signal and wait.
            self._log.debug(
                "Instance seems down; will attempt restart and wait."
            )
            self._do_restart()
            self.wait()
            return

        self._log.debug("Pre-reboot boot_id: %s", pre_boot_id)

        # The instance is reachable, so do the restart and wait for changed
        # boot id
        self._do_restart(**kwargs)
        if wait:
            self.wait_for_restart(old_boot_id=pre_boot_id)

    @abstractmethod
    def _do_restart(self, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def shutdown(self, wait=True, **kwargs):
        """Shutdown the instance.

        Args:
            wait: wait for the instance to shutdown
        """
        raise NotImplementedError

    @abstractmethod
    def start(self, wait=True):
        """Start the instance.

        Args:
            wait: wait for the instance to start.
        """
        raise NotImplementedError

    def _wait_for_instance_start(self):
        """Wait for the cloud instance to be up.

        Subclasses should implement this if their cloud provides a way of
        detecting when an instance has started through their API.
        """

    def wait(self):
        """Wait for instance to be up and cloud-init to be complete."""
        self._wait_for_instance_start()
        self._wait_for_execute()
        self._wait_for_cloudinit()

    def wait_for_restart(self, old_boot_id):
        """Wait for instance to be restarted and cloud-init to be complete.

        old_boot_id is the boot id prior to restart
        """
        self._wait_for_instance_start()
        self._wait_for_execute(old_boot_id=old_boot_id)
        self._wait_for_cloudinit()

    @abstractmethod
    def wait_for_delete(self, **kwargs):
        """Wait for instance to be deleted."""
        raise NotImplementedError

    @abstractmethod
    def wait_for_stop(self):
        """Wait for instance stop."""
        raise NotImplementedError

    def add_network_interface(self) -> str:
        """Add nic to running instance."""
        raise NotImplementedError

    def remove_network_interface(self, ip_address: str):
        """Remove nic from running instance."""
        raise NotImplementedError

    def __del__(self):
        """Cleanup of instance."""
        if self._sftp_client:
            try:
                self._sftp_client.close()
            except SSHException:
                self._log.warning("Failed to close SFTP connection.")
            self._sftp_client = None
        if self._ssh_client:
            try:
                self._ssh_client.close()
            except SSHException:
                self._log.warning("Failed to close SSH connection.")
            self._ssh_client = None

    def clean(self):
        """Clean an instance to make it look prestine.

        This will clean out specifically the cloud-init files and system logs.
        """
        result = self.execute("sudo cloud-init clean --logs --machine-id")
        if result.failed:
            # --machine-id likely isn't supported on this version.
            # Manually reset machine-id even though this is less portable
            self.execute("sudo cloud-init clean --logs")
            self.execute("sudo echo 'uninitialized' > /etc/machine-id")
        self.execute("sudo rm -rf /var/log/syslog")

    def _run_command(self, command, stdin):
        """Run command in the instance."""
        return self._ssh(list(command), stdin=stdin)

    def execute(
        self,
        command,
        stdin=None,
        description=None,
        *,
        use_sudo=False,
        no_log=False,
        **kwargs,
    ):
        """Execute command in instance, recording output, error and exit code.

        Assumes functional networking and execution with the target filesystem
        being available at /.

        Args:
            command: the command to execute as root inside the image. If
                     command is a string, then it will be executed as:
                     `['sh', '-c', command]`
            stdin: bytes content for standard in
            description: purpose of command
            use_sudo: boolean to run the command as sudo

        Returns:
            Result object

        Raises SSHException if there are any problem with the ssh connection

        """
        if isinstance(command, str):
            command = ["sh", "-c", command]
        if use_sudo:
            command = ["sudo", "--"] + command

        if not no_log:
            self._log.info("executing: %s", shell_quote(command))
            if description:
                self._log.debug(description)
            else:
                self._log.debug("executing: %s", shell_quote(command))

        return self._run_command(command, stdin, **kwargs)

    def install(self, packages):
        """Install specific packages.

        Args:
            packages: string or list of package(s) to install

        Returns:
            result from install

        """
        if isinstance(packages, str):
            packages = packages.split(" ")

        self.execute(["sudo", "apt-get", "update"])
        return self.execute(
            [
                "DEBIAN_FRONTEND=noninteractive",
                "sudo",
                "apt-get",
                "install",
                "--yes",
            ]
            + packages
        )

    def pull_file(self, remote_path, local_path):
        """Copy file at 'remote_path', from instance to 'local_path'.

        Args:
            remote_path: path on remote instance
            local_path: local path

        Raises SSHException if there are any problem with the ssh connection
        """
        self._log.debug("pulling file %s to %s", remote_path, local_path)

        sftp = self._sftp_connect()
        sftp.get(remote_path, local_path)

    def push_file(self, local_path, remote_path):
        """Copy file at 'local_path' to instance at 'remote_path'.

        Args:
            local_path: local path
            remote_path: path on remote instance

        Raises SSHException if there are any problem with the ssh connection
        """
        self._log.debug("pushing file %s to %s", local_path, remote_path)

        sftp = self._sftp_connect()
        sftp.put(local_path, remote_path)

    def run_script(self, script, description=None):
        """Run script in target and return stdout.

        Args:
            script: script contents
            description: purpose of script

        Returns:
            result from script execution

        Raises SSHException if there are any problem with the ssh connection
        """
        # Just write to a file, add execute, run it, then remove it.
        shblob = "; ".join(
            (
                "set -e",
                's="$1"',
                "shift",
                'cat > "$s"',
                'trap "rm -f $s" EXIT',
                'chmod +x "$s"',
                '"$s" "$@"',
            )
        )
        return self.execute(
            ["sh", "-c", shblob, "runscript", self._tmpfile()],
            stdin=script,
            description=description,
        )

    def update(self):
        """Run apt-get update/upgrade on instance.

        Returns:
            result from upgrade

        """
        self.execute(["sudo", "apt-get", "update"])
        return self.execute(
            [
                "DEBIAN_FRONTEND=noninteractive",
                "sudo",
                "apt-get",
                "--yes",
                "upgrade",
            ]
        )

    def _ssh(self, command, stdin=None):
        """Run a command via SSH.

        Args:
            command: string or list of the command to run
            stdin: optional, values to be passed in

        Returns:
            tuple of stdout, stderr and the return code

        """
        cmd = shell_pack(command)
        client = self._ssh_connect()
        try:
            fp_in, fp_out, fp_err = client.exec_command(cmd)
        except (ConnectionResetError, NoValidConnectionsError, EOFError) as e:
            raise SSHException from e
        channel = fp_in.channel

        if stdin is not None:
            fp_in.write(stdin)
            fp_in.close()

        channel.shutdown_write()

        out = fp_out.read()
        err = fp_err.read()
        return_code = channel.recv_exit_status()

        out = "" if not out else out.rstrip().decode("utf-8")
        err = "" if not err else err.rstrip().decode("utf-8")

        return Result(out, err, return_code)

    def _ssh_connect(self):
        """Connect to instance via SSH."""
        if self._ssh_client and self._ssh_client.get_transport().is_active():
            return self._ssh_client

        logging.getLogger("paramiko").setLevel(logging.INFO)
        client = paramiko.SSHClient()
        client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

        # Paramiko can barf on valid keys when initializing this way,
        # so the check here is _only_ for checking if we have a
        # password protected keyfile. The filename is passed directly
        # when connecting
        try:
            paramiko.RSAKey.from_private_key_file(
                self.key_pair.private_key_path
            )
        except PasswordRequiredException:
            self._log.warning(
                "The specified key (%s) requires a passphrase. If you have not"
                " added this key to a running SSH agent, you will see failures"
                " to connect after a long timeout.",
                self.key_pair.private_key_path,
            )
        except SSHException:
            pass

        try:
            client.connect(
                username=self.username,
                hostname=self.ip,
                port=int(self.port),
                timeout=self.connect_timeout,
                banner_timeout=self.banner_timeout,
                key_filename=self.key_pair.private_key_path,
            )
        except (
            ConnectionRefusedError,
            AuthenticationException,
            BadHostKeyException,
            ConnectionResetError,
            SSHException,
            OSError,
        ) as e:
            raise SSHException from e
        self._ssh_client = client
        return client

    def _sftp_connect(self):
        """Connect to instance via SFTP."""
        if (
            self._sftp_client
            and self._sftp_client.get_channel().get_transport().is_active()
        ):
            return self._sftp_client

        logging.getLogger("paramiko").setLevel(logging.INFO)

        # _ssh_connect() implements the required retry logic.
        client = self._ssh_connect()
        sftpclient = client.open_sftp()
        self._sftp_client = sftpclient
        return sftpclient

    def _tmpfile(self):
        """Get a tmp file in the target.

        Returns:
            path to new file in target

        """
        path = "/tmp/%s-%04d" % (type(self).__name__, self._tmp_count)
        self._tmp_count += 1
        return path

    def _wait_for_execute(self, old_boot_id=None):
        """Wait until we can execute a command in the instance.

        If old_boot_id is specified, we use its value to wait until we
        find a new boot id
        """
        self._log.info("_wait_for_execute to complete")

        # Wait 40 minutes before failing. AWS EC2 metal instances can take
        # over 20 minutes to start or restart, so we shouldn't lower
        # this timeout
        start = time.time()
        end = start + 40 * 60
        while time.time() < end:
            with suppress(SSHException, OSError):
                boot_id = self.get_boot_id()
                if not old_boot_id or boot_id != old_boot_id:
                    return
            time.sleep(1)

        raise PycloudlibTimeoutError(
            "Instance can't be reached after 40 minutes. "
            "Failed to obtain new boot id",
        )

    def _wait_for_cloudinit(self):
        """Wait until cloud-init has finished."""
        self._log.info("_wait_for_cloudinit to complete")
        if self.execute("command -v systemctl").ok:
            # We may have issues with cloud-init status early boot, so also
            # ensure our cloud-init.target is active as an extra layer of
            # protection against connecting before the system is ready
            for _ in range(300):
                with suppress(SSHException):
                    if self.execute(
                        ["systemctl", "is-active", "cloud-init.target"],
                        no_log=True,
                    ).ok:
                        break
                time.sleep(1)
        cmd = ["cloud-init", "status", "--wait", "--long"]
        self.execute(cmd, description="waiting for start")

    def _sync_filesystem(self):
        """Sync the filesystem before powering down."""
        with suppress(SSHException):
            self.execute("sync")



# This file is part of pycloudlib. See LICENSE file for license information.
"""Utilities for OCI images and instances."""
import time
from typing import TYPE_CHECKING

from pycloudlib.errors import PycloudlibError

if TYPE_CHECKING:
    import oci


def wait_till_ready(func, current_data, desired_state, sleep_seconds=1000):
    """Wait until the results of function call reach a desired lifecycle state.

    Args:
        func: The function to call
        current_data: Structure containing the initial id and lifecycle state
        desired_state: Desired value of "lifecycle_state"
        sleep_seconds: How long to wait in seconds
    Returns:
        The updated version of the current_data
    """
    for _ in range(sleep_seconds):
        current_data = func(current_data.id).data
        if current_data.lifecycle_state == desired_state:
            return current_data
        time.sleep(1)
    raise PycloudlibError(
        "Expected {} state, but found {} after waiting {} seconds. "
        "Check OCI console for more details".format(
            desired_state, current_data.lifecycle_state, sleep_seconds
        )
    )


def get_subnet_id(
    network_client: "oci.core.VirtualNetworkClient",  # type: ignore
    compartment_id: str,
    availability_domain: str,
) -> str:
    """Get a subnet id linked to `availability_domain`.

    From specified compartment select the first subnet linked to
    `availability_domain` or the first one.

    Args:
        network_client: Instance of VirtualNetworkClient.
        compartment_id: Compartment where the subnet has to belong
        availability_domain: Domain to look for subnet id in.
    Returns:
        The updated version of the current_data
    Raises:
        `Exception` if unable to determine `subnet_id` for
        `availability_domain`
    """
    vcn_id = network_client.list_vcns(compartment_id).data[0].id
    subnets = network_client.list_subnets(compartment_id, vcn_id=vcn_id).data
    subnet_id = None
    for subnet in subnets:
        if subnet.availability_domain == availability_domain:
            subnet_id = subnet.id
            break
    else:
        subnet_id = subnets[0].id
    if not subnet_id:
        raise PycloudlibError(
            f"Unable to determine subnet id for domain: {availability_domain}"
        )
    return subnet_id



"""Tests related to pycloudlib.gce.cloud module."""
import mock
import pytest

from pycloudlib.cloud import ImageType
from pycloudlib.gce.cloud import GCE

# mock module path
MPATH = "pycloudlib.gce.cloud."


class FakeGCE(GCE):
    """GCE Class that doesn't load config or make requests during __init__."""

    # pylint: disable=super-init-not-called
    def __init__(self, *_, **__):
        """Fake __init__ that sets mocks for needed variables."""
        self._log = mock.MagicMock()
        self.compute = mock.MagicMock()


# pylint: disable=protected-access,missing-function-docstring
class TestGCE:
    """General GCE testing."""

    @pytest.mark.parametrize(
        [
            "release",
            "arch",
            "api_side_effects",
            "expected_filter_calls",
            "expected_image_list",
        ],
        [
            pytest.param(
                "xenial",
                "arm64",
                Exception(),
                [],
                [],
                id="xenial_no_arm64_support_zero_sdk_list_calls",
            ),
            pytest.param(
                "xenial",
                "x86_64",
                [{"items": [1, 2, 3]}, Exception()],
                [
                    mock.call(
                        project="project-name",
                        filter="name=name-filter",
                        maxResults=500,
                        pageToken="",
                    )
                ],
                [1, 2, 3],
                id="xenial_x86_64_suppport_one_sdk_list_call_empty_pagetoken",
            ),
            pytest.param(
                "kinetic",
                "arm64",
                [
                    {"items": [1, 2, 3], "nextPageToken": "something"},
                    {"items": [4, 5, 6]},
                    Exception(),
                ],
                [
                    mock.call(
                        project="project-name",
                        filter="(name=name-filter) AND (architecture=ARM64)",
                        maxResults=500,
                        pageToken="",
                    ),
                    mock.call(
                        project="project-name",
                        filter="(name=name-filter) AND (architecture=ARM64)",
                        maxResults=500,
                        pageToken="something",
                    ),
                ],
                [1, 2, 3, 4, 5, 6],
                id="non_xenial_arm64_suppport_one_sdk_list_call_per_page",
            ),
        ],
    )
    def test_query_image_list(  # noqa: D102
        self,
        release,
        arch,
        api_side_effects,
        expected_filter_calls,
        expected_image_list,
    ):
        gce = FakeGCE(tag="tag")
        with mock.patch.object(gce, "compute") as m_compute:
            m_execute = mock.MagicMock(
                name="m_execute", side_effect=api_side_effects
            )
            m_executor = mock.MagicMock(name="m_executor")
            m_executor.execute = m_execute
            m_list = mock.MagicMock(name="m_list", return_value=m_executor)
            m_lister = mock.MagicMock(name="m_lister")
            m_lister.list = m_list
            m_images = mock.MagicMock(name="m_images", return_value=m_lister)
            m_compute.images = m_images

            assert expected_image_list == gce._query_image_list(
                release, "project-name", "name-filter", arch
            )
            assert m_list.call_args_list == expected_filter_calls

    @mock.patch(
        MPATH + "GCE._query_image_list",
        return_value=[
            {
                "id": "2",
                "name": "2",
                "creationTimestamp": "2",
            },
            {
                "id": "4",
                "name": "4",
                "creationTimestamp": "4",
            },
            {
                "id": "1",
                "name": "1",
                "creationTimestamp": "1",
            },
            {
                "id": "3",
                "name": "3",
                "creationTimestamp": "3",
            },
        ],
    )
    @mock.patch(MPATH + "GCE._get_name_filter", return_value="name-filter")
    @mock.patch(MPATH + "GCE._get_project", return_value="project-name")
    def test_daily_image_returns_latest_from_query(  # noqa: D102
        self,
        m_get_project,
        m_get_name_filter,
        m_query_image_list,
    ):
        gce = FakeGCE(tag="tag")
        image = gce.daily_image(
            "jammy", arch="x86_64", image_type=ImageType.GENERIC
        )
        assert m_get_project.call_args_list == [
            mock.call(image_type=ImageType.GENERIC)
        ]
        assert m_get_name_filter.call_args_list == [
            mock.call(release="jammy", image_type=ImageType.GENERIC)
        ]
        assert m_query_image_list.call_args_list == [
            mock.call("jammy", "project-name", "name-filter", "x86_64")
        ]
        assert image == "projects/project-name/global/images/4"



# This file is part of pycloudlib. See LICENSE file for license information.
"""LXD default values to be used by cloud and instance modules."""
import textwrap

LXC_PROFILE_VERSION = "v3"


# For Xenial and Bionic vendor-data required to setup lxd-agent in a vm
LXC_SETUP_VENDORDATA = textwrap.dedent(
    """\
    config:
      user.vendor-data: |
        #cloud-config
        write_files:
        - path: /var/lib/cloud/scripts/per-once/setup-lxc.sh
          encoding: b64
          permissions: '0755'
          owner: root:root
          content: |
              IyEvYmluL3NoCmlmICEgZ3JlcCBseGRfY29uZmlnIC9wcm9jL21vdW50czsgdGhlbgogICAgbWtk
              aXIgLXAgL3J1bi9seGRhZ2VudAogICAgbW91bnQgLXQgOXAgY29uZmlnIC9ydW4vbHhkYWdlbnQK
              ICAgIFZJUlQ9JChzeXN0ZW1kLWRldGVjdC12aXJ0KQogICAgY2FzZSAkVklSVCBpbgogICAgICAg
              IHFlbXV8a3ZtKQogICAgICAgICAgICAoY2QgL3J1bi9seGRhZ2VudC8gJiYgLi9pbnN0YWxsLnNo
              KQogICAgICAgICAgICB1bW91bnQgL3J1bi9seGRhZ2VudAogICAgICAgICAgICBzeXN0ZW1jdGwg
              c3RhcnQgbHhkLWFnZW50CiAgICAgICAgICAgIDs7CiAgICAgICAgKikKICAgIGVzYWMKZmkK
   """
)

VM_PROFILE_TMPL = textwrap.dedent(
    """\
    {vendordata}
    description: Pycloudlib LXD profile for {series} VMs
    devices:
      {config_device}
      eth0:
        name: eth0
        network: lxdbr0
        type: nic
      root:
        path: /
        pool: default
        type: disk
    name: vm
    """
)


def _make_vm_profile(
    series: str, *, install_agent: bool, config_cloudinit: bool
) -> str:
    config_device = ""
    vendordata = "config: {}"
    if config_cloudinit:
        # We need to mount the config drive so that cloud-init finds the
        # vendor-data instructing it to install the agent
        config_device = "config: {source: cloud-init:config, type: disk}"
    if install_agent:
        vendordata = LXC_SETUP_VENDORDATA
    return VM_PROFILE_TMPL.format(
        config_device=config_device, series=series, vendordata=vendordata
    )


base_vm_profiles = {
    "xenial": _make_vm_profile(
        "xenial", install_agent=False, config_cloudinit=True
    ),
    "bionic": _make_vm_profile(
        "bionic", install_agent=True, config_cloudinit=True
    ),
    "default": _make_vm_profile(
        "default", install_agent=False, config_cloudinit=False
    ),
}



# This file is part of pycloudlib. See LICENSE file for license information.
"""Azure Util Functions."""
import logging
import re

from azure.core.exceptions import ClientAuthenticationError
from azure.identity import AzureCliCredential, ClientSecretCredential
from knack.util import CLIError

from pycloudlib import util
from pycloudlib.errors import CloudSetupError

logger = logging.getLogger(__name__)

RE_AZURE_IMAGE_ID = (
    r"(?P<publisher>[^:]+):(?P<offer>[^:]+):(?P<sku>[^:]+)(:(?P<version>.*))?"
)


def get_client(resource, config_dict: dict):
    """Get azure client based on the give resource.

    This method will first verify if we can get the client
    by using the information provided on the login account
    of the user machine. If the user is not logged into Azure,
    we will try to get the client from the ids given by the
    user to this class.

    Args:
        resource: Azure Resource, An Azure resource that we want to get
                  a client for.
        config_dict: dict, Id parameters passed by the user to this class.

    Returns:
        The client for the resource passed as parameter.

    """
    if not util.subp("command -v az", shell=True, rcs=[0, 1, 127]).ok:
        logger.debug(
            "No azure-cli 'az' command found. Trying explicit config params"
        )
    else:
        try:
            cli_credential = AzureCliCredential()
            subscription_id = config_dict.get("subscriptionId")
            client = resource(cli_credential, subscription_id=subscription_id)
            return client
        except CLIError:
            logger.debug(
                "No valid azure-cli config found."
                " Trying explicit config params"
            )
        except ClientAuthenticationError:
            logger.debug(
                "Authentication error: No valid azure-cli config found."
                " Trying explicit config params"
            )

    required_keys = frozenset(
        {"clientId", "clientSecret", "tenantId", "subscriptionId"}
    )
    missing_keys = required_keys.difference(set(config_dict.keys()))
    if missing_keys:
        raise CloudSetupError(
            "No AZ cli config found, missing required keys: {}".format(
                ", ".join(missing_keys)
            )
        )

    parameters = {
        "active_directory": "https://login.microsoftonline.com",
        "resource_manager": "https://management.azure.com/",
        "active_directory_graph_resource_id": "https://graph.windows.net/",
        "sql_management": "https://management.core.windows.net:8443/",
        "gallery": "https://gallery.azure.com/",
        "management": "https://management.core.windows.net/",
    }
    parameters.update(config_dict)
    credential = ClientSecretCredential(
        tenant_id=parameters["tenantId"],
        client_id=parameters["clientId"],
        client_secret=parameters["clientSecret"],
        authority=parameters["active_directory"],
    )

    return resource(
        credential, subscription_id=parameters["subscriptionId"], **parameters
    )


def parse_image_id(image_id):
    """Extract publisher, offer, sku and optional version from image_id.

    The image_id is expected to be a string in the following
    format: Canonical:UbuntuServer:19.10-DAILY[:latest]

    Args:
        image_id: string, The image id

    Returns
        Dict with publisher, offer and sku and optional version keys.

    """
    match = re.match(RE_AZURE_IMAGE_ID, image_id)
    if not match:
        # Snapshot image ids do not follow the publisher:offer:sku pattern
        return {}

    return match.groupdict()


def get_resource_group_name_from_id(resource_id):
    """Retrieve the resource group name of a resource.

    Args:
        resource_id: string, the resource id

    Returns:
        A string representing the resource group

    """
    return resource_id.split("/")[4]


def get_resource_name_from_id(resource_id):
    """Retrieve the name of a resource.

    Args:
        resource_id: string, the resource id

    Returns:
        A string representing the resource name

    """
    return resource_id.split("/")[-1]


def get_image_reference_params(image_id):
    """Return the correct parameter for image reference based on image id.

    Verify if the image id is associated with a current image found
    on Azure Marketplace or a custom image, for example, created through
    a snapshot process. Depending on the image id format, we can
    differentiate if we should create image parameters for a
    Marketplace image or a custom image.

    Args:
        image_id: string, Represents a image to be used when provisioning
                  a virtual machine

    Returns:
        A dict representing the image reference parameters that will be
        used to provision a virtual machine

    """
    img_dict = parse_image_id(image_id)
    if img_dict.get("publisher") == "Canonical":
        # If the image id starts with 'Canonical", we know that it is a
        # marketplace image, and to we must reference it using the
        # combination of publisher, offer, sku and version info
        img_dict.update({"version": "latest"})
        return img_dict

    # Custom images can be directly referenced by their id
    return {"id": image_id}


def is_pro_image(image_id, registered_image):
    """Verify if the image id represents a pro image.

    Check the image id string for patterns found only on
    pro images. However, snapshot images do not have pro
    information on ther image id. We are enconding that
    information on the registed_image dict, which represents
    the base image that created the snapshot. Therefore,
    we fail at looking in the image id string, we look it up
    at the registered_image dict.

    Args:
        image_id: string, Represents a image to be used when provisioning
                  a virtual machine
        registered_image: dict, Represents the base image used for creating
                          the image referenced by image_id. This will only
                          happen for snapshot images.

    Returns:
        A boolean indicating if the image is pro image

    """
    offer = ""
    img_dict = parse_image_id(image_id)
    if img_dict.get("publisher") == "Canonical":
        offer = img_dict["offer"]
    elif registered_image is not None:
        offer = registered_image["offer"] or ""

    return bool("-pro-" in offer)


def get_plan_params(image_id, registered_image):
    """Return the correct parameter for plan based on pro image id.

    Args:
        image_id: string, Represents a image to be used when provisioning
                  a virtual machine
        registered_image: dict, Represents the base image used for creating
                          the image referenced by image_id. This will only
                          happen for snapshot images.

    Returns:
        A dict representing the plan parameters that will be
        used to provision a virtual machine

    """
    if registered_image is not None:
        return {
            "name": registered_image["sku"],
            "product": registered_image["offer"],
            "publisher": "canonical",
        }

    img_dict = parse_image_id(image_id)
    return {
        "name": img_dict.get("sku"),
        "product": img_dict.get("offer"),
        "publisher": img_dict.get("publisher", "").lower(),
    }



"""Test the errors.py module."""
import pytest

from pycloudlib.errors import ResourceNotFoundError, ResourceType


class TestResourceType:
    """Tests related to `ResourceType`."""

    @pytest.mark.parametrize(
        "item",
        list(map(lambda item: pytest.param(item, id=item.name), ResourceType)),
    )
    def test_str_representable(self, item):
        """Test that all instances of `ResourceType` are convertible to str."""
        assert str(item)


class TestResourceNotFoundError:
    """Tests related to `ResourceNotFoundError`."""

    @pytest.mark.parametrize(
        ["exception", "expected_msg"],
        [
            (
                ResourceNotFoundError(
                    resource_type=ResourceType.INSTANCE,
                    resource_id="id",
                    resource_name="name",
                    custom_key="custom_key",
                ),
                (
                    "Could not locate the resource type `instance`: "
                    "id=id, name=name, custom_key=custom_key"
                ),
            ),
            (
                ResourceNotFoundError(
                    resource_type=ResourceType.IMAGE,
                    resource_id="id",
                    custom_key="custom_key",
                ),
                (
                    "Could not locate the resource type `image`: "
                    "id=id, custom_key=custom_key"
                ),
            ),
            (
                ResourceNotFoundError(
                    resource_type=ResourceType.NETWORK,
                    resource_name="name",
                    custom_key="custom_key",
                ),
                (
                    "Could not locate the resource type `network`: "
                    "name=name, custom_key=custom_key"
                ),
            ),
            (
                ResourceNotFoundError(
                    resource_type=ResourceType.NETWORK,
                ),
                ("Could not locate the resource type `network`"),
            ),
        ],
    )
    def test_exception_message(self, exception, expected_msg):
        """Test that exceptions have correct error messages."""
        assert expected_msg == str(exception)



"""Module containing pycloudlib errors.

Each cloud can have specific errors, please refer to each
`pycloudlib.<cloud>.errors` module.
"""
import enum
from typing import Optional


class PycloudlibException(Exception):
    """Root pycloudlib exception.

    This exception is not meant to be raised by pycloudlib. The intention
    is that every custom pycloudlib exception will inherit from this one,
    allowing client code to catch any expection by catching this one.
    """


class PycloudlibError(PycloudlibException):
    """Error that doesn’t fall in any of the other categories."""


class ResourceType(enum.Enum):
    """Represent types of resources."""

    IMAGE = enum.auto()
    INSTANCE = enum.auto()
    NETWORK = enum.auto()

    def __str__(self) -> str:  # noqa: D105
        if self == self.INSTANCE:
            return "instance"
        if self == self.IMAGE:
            return "image"
        if self == self.NETWORK:
            return "network"
        raise NotImplementedError


class ResourceNotFoundError(PycloudlibException):
    """Raised when a resource is not found.

    Examples:
    ---------
    >>> e = ResourceNotFoundError(ResourceType.IMAGE, "id-123")
    >>> e.resource_id
    'id-123'
    >>> raise e  # doctest: +ELLIPSIS
    Traceback (most recent call last):
        ...
    pycloudlib.errors.ResourceNotFoundError: \
Could not locate the resource type `image`: id=id-123
    """

    def __init__(
        self,
        resource_type: ResourceType,
        resource_id: Optional[str] = None,
        resource_name: Optional[str] = None,
        **kwargs,
    ):
        """Init method.

        :param resource_type: Instance of `ResourceType`
        :param resource_id: Resource's id
        :param resource_type: Resource's name
        """
        super().__init__()
        self.resource_type = resource_type
        self.resource_id = resource_id
        self.resource_name = resource_name
        self._extra_info = kwargs

    def __str__(self) -> str:  # noqa: D105
        resource_info = self.__render_resource()
        msg = f"Could not locate the resource type `{self.resource_type}`"
        if resource_info:
            msg += f": {resource_info}"
        return msg

    def __render_resource(self) -> str:
        parts = []
        if self.resource_id:
            parts.append(f"id={self.resource_id}")
        if self.resource_name:
            parts.append(f"name={self.resource_name}")
        if self._extra_info:
            parts.extend(
                map(
                    lambda item: f"{item[0]}={item[1]}",
                    self._extra_info.items(),
                )
            )
        return ", ".join(parts)


class ImageNotFoundError(ResourceNotFoundError):
    """Sepecialized's `ResourceNotFoundError` for images."""

    def __init__(self, *args, **kwargs):  # noqa: D107
        super().__init__(ResourceType.IMAGE, *args, **kwargs)


class InstanceNotFoundError(ResourceNotFoundError):
    """Sepecialized's `ResourceNotFoundError` for instances."""

    def __init__(self, *args, **kwargs):  # noqa: D107
        super().__init__(ResourceType.INSTANCE, *args, **kwargs)


class NetworkNotFoundError(ResourceNotFoundError):
    """Sepecialized's `ResourceNotFoundError` for networks."""

    def __init__(self, *args, **kwargs):  # noqa: D107
        super().__init__(ResourceType.NETWORK, *args, **kwargs)


class CloudSetupError(PycloudlibException):
    """Raised if there is some problem with a pycloudlib's Cloud set up."""


class CloudError(PycloudlibException):
    """Represents errors comming from Cloud's SDKs."""


class PycloudlibTimeoutError(PycloudlibException):
    """Timeout error."""



# This file is part of pycloudlib. See LICENSE file for license information.
"""GCE instance."""

from time import sleep

import googleapiclient.discovery
from googleapiclient.errors import HttpError

from pycloudlib.errors import PycloudlibTimeoutError
from pycloudlib.gce.util import get_credentials, raise_on_error
from pycloudlib.instance import BaseInstance


class GceInstance(BaseInstance):
    """GCE backed instance."""

    _type = "gce"

    def __init__(
        self,
        key_pair,
        instance_id,
        project,
        zone,
        credentials_path,
        *,
        name=None,
    ):
        """Set up the instance.

        Args:
            key_pair: A KeyPair for SSH interactions
            instance_id: Id returned when creating the instance
            project: Project instance was created in
            zone: Zone instance was created in
        """
        if project is None or zone is None:
            raise ValueError(
                "kwargs 'project' and 'zone' are required. "
                "Project: {}, Zone: {}".format(project, zone)
            )
        super().__init__(key_pair)
        self.instance_id = instance_id
        self._name = name
        self.project = project
        self.zone = zone
        self._ip = None
        credentials = get_credentials(credentials_path)
        self.instance = googleapiclient.discovery.build(
            "compute", "v1", cache_discovery=False, credentials=credentials
        ).instances()

    def __repr__(self):
        """Create string representation of class."""
        return "{}(instance_id={})".format(
            self.__class__.__name__,
            self.instance_id,
        )

    @property
    def id(self):
        """Return the instance id."""
        return self.instance_id

    @property
    def name(self):
        """Return the instance name."""
        if not self._name:
            result = self.instance.get(
                project=self.project,
                zone=self.zone,
                instance=self.instance_id,
            ).execute()
            self._name = result["name"]
        return self._name

    @property
    def ip(self):
        """Return IP address of instance."""
        if not self._ip:
            self._ip = self._get_ip()
        return self._ip

    def _get_ip(self):
        result = self.instance.get(
            project=self.project,
            zone=self.zone,
            instance=self.instance_id,
        ).execute()
        ip = result["networkInterfaces"][0]["accessConfigs"][0]["natIP"]
        return ip

    def delete(self, wait=True):
        """Delete the instance.

        Args:
            wait: wait for instance to be deleted
        """
        response = self.instance.delete(
            project=self.project, zone=self.zone, instance=self.instance_id
        ).execute()
        raise_on_error(response)
        if wait:
            self.wait_for_delete()

    def _do_restart(self, **kwargs):
        """Restart the instance."""
        self.shutdown(wait=True)
        self.start(wait=False)

    def shutdown(self, wait=True, **kwargs):
        """Shutdown the instance.

        Args:
            wait: wait for the instance to shutdown
        """
        response = self.instance.stop(
            project=self.project, zone=self.zone, instance=self.instance_id
        ).execute()
        raise_on_error(response)
        if wait:
            self.wait_for_stop()

    def start(self, wait=True):
        """Start the instance.

        Args:
            wait: wait for the instance to start.
        """
        response = self.instance.start(
            project=self.project, zone=self.zone, instance=self.instance_id
        ).execute()
        raise_on_error(response)
        if wait:
            self.wait()

    def _wait_for_instance_start(self):
        """Wait for instance to be up."""
        self._wait_for_status("RUNNING")
        self._ip = self._get_ip()

    def wait_for_delete(self, sleep_seconds=30, raise_on_fail=False):
        """Wait for instance to be deleted."""
        for _ in range(sleep_seconds):
            try:
                result = self.instance.get(
                    project=self.project,
                    zone=self.zone,
                    instance=self.instance_id,
                ).execute()
                if result["status"] == "TERMINATED":
                    break
            except HttpError as e:
                # Sometimes URL just 404s once deleted
                if e.resp.status == 404:
                    break
                raise e
        else:
            msg = (
                f"Instance not terminated after {sleep_seconds} seconds. "
                "Check GCE console."
            )
            if raise_on_fail:
                raise PycloudlibTimeoutError(msg)
            self._log.warning(msg)

    def wait_for_stop(self):
        """Wait for instance stop."""
        self._wait_for_status("TERMINATED")

    def _wait_for_status(self, status, sleep_seconds=300):
        response = {"status": None}
        for _ in range(sleep_seconds):
            response = self.instance.get(
                project=self.project, zone=self.zone, instance=self.instance_id
            ).execute()
            if response["status"] == status:
                break
            sleep(1)
        else:
            raise PycloudlibTimeoutError(
                f"Expected {status} state, but found {response['status']} "
                f"after waiting {sleep_seconds} seconds. "
                "Check GCE console for more details."
            )



"""Tests related to lxd._images."""
import json
from unittest import mock

import pytest

from pycloudlib.lxd import _images

M_PATH = "pycloudlib.lxd._images."


@mock.patch(M_PATH + "subp")
class TestImages:  # pylint: disable=W0212
    """Test class for _images."""

    @pytest.mark.parametrize(
        "images",
        (
            [],
            [{"properties": {"serial": "s0"}}],
            [
                {"properties": {"serial": "s0"}},
                {"properties": {"serial": "s1"}},
            ],
        ),
    )
    @mock.patch(M_PATH + "_find_images")
    def test_find_image_serial(self, m_find_images, m_subp, images):
        """Test find_image_serial method."""
        image_id = "my:image_id"
        mock.Mock()
        if not images:
            expected_result = None
        else:
            expected_result = "s0"

        m_find_images.return_value = images
        assert expected_result == _images.find_image_serial(image_id)

        assert [
            mock.call("my", (("fingerprint", "image_id"),))
        ] == m_find_images.call_args_list
        assert [] == m_subp.call_args_list

    @pytest.mark.parametrize(
        ["filters", "expected_call"],
        (
            (
                None,
                mock.call(
                    [
                        "lxc",
                        "image",
                        "list",
                        "remote:",
                        "--format=json",
                    ]
                ),
            ),
            (
                (("filter_0", "value_0"), ("f1", "v1")),
                mock.call(
                    [
                        "lxc",
                        "image",
                        "list",
                        "remote:",
                        "--format=json",
                        "filter_0=value_0",
                        "f1=v1",
                    ]
                ),
            ),
        ),
    )
    def test_find_images(self, m_subp, filters, expected_call):
        """Test find_images method."""
        content = ["image_0", "image_1"]
        m_subp.return_value = json.dumps(content)
        remote = "remote"
        assert content == _images._find_images(remote, filters)
        assert [expected_call] == m_subp.call_args_list

    @pytest.mark.parametrize(
        ["remote_in", "remote_out"],
        (
            (None, "ubuntu-daily:"),
            ("remote", "remote:"),
            ("remote:", "remote:"),
            ("remote:fingerprint", "remote:fingerprint"),
        ),
    )
    def test_normalize_remote(self, m_subp, remote_in, remote_out):
        """Test _normalize_remote method."""
        assert remote_out == _images._normalize_remote(remote_in)
        assert [] == m_subp.call_args_list

    @pytest.mark.parametrize(
        ["images", "output"],
        (
            ([{"properties": {"os": "ubuntu", "release": "lunar"}}], "lunar"),
            ([{"properties": {"os": "Ubuntu", "release": "lunar"}}], "lunar"),
            ([], None),
            ([{"x": {"os": "ubuntu", "release": "lunar"}}], None),
            ([{"properties": {"x": "ubuntu", "release": "lunar"}}], None),
            ([{"properties": {"os": "ubuntu", "x": "lunar"}}], None),
        ),
    )
    @mock.patch(M_PATH + "_find_images")
    def test_find_release(self, m_find_images, m_subp, images, output):
        """Test find_release method."""
        m_find_images.return_value = images
        assert output == _images.find_release("remote:image_id")
        assert [] == m_subp.call_args_list

    @pytest.mark.parametrize(
        ["images", "n_find_images_calls", "is_container", "expected_output"],
        (
            (
                [{"fingerprint": "asdf", "type": "container"}],
                1,
                True,
                "ubuntu-daily:asdf",
            ),
            (
                [{"fingerprint": "asdf", "type": "squashfs"}],
                2,
                True,
                "ubuntu-daily:asdf",
            ),
            (
                [{"fingerprint": "asdf", "type": "unknown"}],
                2,
                True,
                None,
            ),
            (
                [{"fingerprint": "asdf", "type": "virtual-machine"}],
                1,
                False,
                "ubuntu-daily:asdf",
            ),
            (
                [{"fingerprint": "asdf", "type": "disk-kvm.img"}],
                2,
                False,
                "ubuntu-daily:asdf",
            ),
            (
                [{"fingerprint": "asdf", "type": "disk1.img"}],
                3,
                False,
                "ubuntu-daily:asdf",
            ),
            (
                [{"fingerprint": "asdf", "type": "uefi1.img"}],
                4,
                False,
                "ubuntu-daily:asdf",
            ),
            (
                [{"fingerprint": "asdf", "type": "unknown"}],
                4,
                False,
                None,
            ),
        ),
    )
    @mock.patch(M_PATH + "_find_images")
    def test_find_last_fingerprint(
        self,
        m_find_images,
        m_subp,
        images,
        n_find_images_calls,
        is_container,
        expected_output,
    ):
        """Test find_last_fingerprint method."""

        def find_images(remote, filters=None):
            # pylint: disable=unused-argument
            filters_map = dict(filters)
            return list(
                filter(lambda img: img["type"] == filters_map["type"], images)
            )

        m_find_images.side_effect = find_images

        daily = True
        release = "bionic"
        arch = "amd64"
        expected_remote = "ubuntu-daily:"

        expected_find_images_calls = [
            mock.call(
                expected_remote,
                (
                    ("architecture", arch),
                    ("release", release),
                    ("label", "daily"),
                    (
                        "type",
                        "container" if is_container else "virtual-machine",
                    ),
                ),
            )
        ]
        if is_container:
            if n_find_images_calls >= 2:
                expected_find_images_calls.append(
                    mock.call(
                        expected_remote,
                        (
                            ("architecture", arch),
                            ("release", release),
                            ("label", "daily"),
                            ("type", "squashfs"),
                        ),
                    )
                )
        else:
            if n_find_images_calls >= 2:
                expected_find_images_calls.append(
                    mock.call(
                        expected_remote,
                        (
                            ("architecture", arch),
                            ("release", release),
                            ("label", "daily"),
                            ("type", "disk-kvm.img"),
                        ),
                    )
                )
            if n_find_images_calls >= 3:
                expected_find_images_calls.append(
                    mock.call(
                        expected_remote,
                        (
                            ("architecture", arch),
                            ("release", release),
                            ("label", "daily"),
                            ("type", "disk1.img"),
                        ),
                    )
                )
            if n_find_images_calls >= 4:
                expected_find_images_calls.append(
                    mock.call(
                        expected_remote,
                        (
                            ("architecture", arch),
                            ("release", release),
                            ("label", "daily"),
                            ("type", "uefi1.img"),
                        ),
                    )
                )

        assert expected_output == _images.find_last_fingerprint(
            daily, release, is_container, arch
        )
        assert expected_find_images_calls == m_find_images.call_args_list
        assert [] == m_subp.call_args_list



# This file is part of pycloudlib. See LICENSE file for license information.
"""IBM's __init__."""
from pycloudlib.ibm.errors import IBMException
from pycloudlib.ibm.instance import VPC, IBMInstance

__all__ = ["IBMException", "IBMInstance", "VPC"]



# This file is part of pycloudlib. See LICENSE file for license information.
"""AWS EC2 Cloud type."""
import re
from typing import Optional

import botocore

from pycloudlib.cloud import BaseCloud, ImageType
from pycloudlib.config import ConfigFile
from pycloudlib.ec2.instance import EC2Instance
from pycloudlib.ec2.util import _get_session, _tag_resource
from pycloudlib.ec2.vpc import VPC
from pycloudlib.errors import (
    CloudSetupError,
    ImageNotFoundError,
    PycloudlibError,
)
from pycloudlib.util import LTS_RELEASES, UBUNTU_RELEASE_VERSION_MAP


class EC2(BaseCloud):
    """EC2 Cloud Class."""

    _type = "ec2"

    def __init__(
        self,
        tag: str,
        timestamp_suffix: bool = True,
        config_file: Optional[ConfigFile] = None,
        *,
        access_key_id: Optional[str] = None,
        secret_access_key: Optional[str] = None,
        region: Optional[str] = None,
    ):
        """Initialize the connection to EC2.

        boto3 will read a users /home/$USER/.aws/* files if no
        arguments are provided here to find values.

        Args:
            tag: string used to name and tag resources with
            timestamp_suffix: bool set True to append a timestamp suffix to the
                tag
            config_file: path to pycloudlib configuration file
            access_key_id: user's access key ID
            secret_access_key: user's secret access key
            region: region to login to
        """
        super().__init__(
            tag,
            timestamp_suffix,
            config_file,
            required_values=[access_key_id, secret_access_key, region],
        )
        self._log.debug("logging into EC2")

        try:
            session = _get_session(
                access_key_id or self.config.get("access_key_id"),
                secret_access_key or self.config.get("secret_access_key"),
                region or self.config.get("region"),
            )
            self.client = session.client("ec2")
            self.resource = session.resource("ec2")
            self.region = session.region_name
        except botocore.exceptions.NoRegionError as e:
            raise CloudSetupError(
                "Please configure default region in $HOME/.aws/config"
            ) from e
        except botocore.exceptions.NoCredentialsError as e:
            raise CloudSetupError(
                "Please configure ec2 credentials in $HOME/.aws/credentials"
            ) from e

    def get_or_create_vpc(self, name, ipv4_cidr="192.168.1.0/20"):
        """Create a or return matching VPC.

        This can be used instead of using the default VPC to create
        a custom VPC for usage.

        Args:
            name: name of the VPC
            ipv4_cidr: CIDR of IPV4 subnet

        Returns:
            VPC object

        """
        # Check to see if current VPC exists
        vpcs = self.client.describe_vpcs(
            Filters=[{"Name": "tag:Name", "Values": [name]}]
        )["Vpcs"]
        if vpcs:
            return VPC.from_existing(self.resource, vpc_id=vpcs[0]["VpcId"])
        return VPC.create(self.resource, name=name, ipv4_cidr=ipv4_cidr)

    def released_image(
        self,
        release: str,
        *,
        arch: str = "x86_64",
        image_type: ImageType = ImageType.GENERIC,
        **kwargs,
    ):
        """Find the id of the latest released image for a particular release.

        Args:
            release: string, Ubuntu release to look for
            arch: string, architecture to use
            root_store: string, root store to use

        Returns:
            string, id of latest image

        """
        self._log.debug("finding released Ubuntu image for %s", release)
        image = self._find_latest_image(
            release=release, arch=arch, image_type=image_type, daily=False
        )
        return image["ImageId"]

    def _get_name_for_image_type(
        self, release: str, image_type: ImageType, daily: bool
    ):
        if image_type == ImageType.GENERIC:
            base_location = "ubuntu/{}/hvm-ssd".format(
                "images-testing" if daily else "images"
            )
            if release in LTS_RELEASES:
                return "{}/ubuntu-{}{}-*-server-*".format(
                    base_location, release, "-daily" if daily else ""
                )

            return "{}/ubuntu-{}{}-*".format(
                base_location, release, "-daily" if daily else ""
            )

        if image_type == ImageType.PRO:
            return "ubuntu-pro-server/images/hvm-ssd/ubuntu-{}-{}-*".format(
                release, UBUNTU_RELEASE_VERSION_MAP[release]
            )

        if image_type == ImageType.PRO_FIPS:
            return "ubuntu-pro-fips*/images/hvm-ssd/ubuntu-{}-{}-*".format(
                release, UBUNTU_RELEASE_VERSION_MAP[release]
            )

        raise ValueError("Invalid image_type")

    def _get_owner(self, image_type: ImageType):
        return (
            "099720109477"
            if image_type == ImageType.GENERIC
            else "aws-marketplace"
        )

    def _get_search_filters(
        self, release: str, arch: str, image_type: ImageType, daily: bool
    ):
        return [
            {
                "Name": "name",
                "Values": [
                    self._get_name_for_image_type(release, image_type, daily)
                ],
            },
            {
                "Name": "architecture",
                "Values": [arch],
            },
        ]

    def _find_latest_image(
        self, release: str, arch: str, image_type: ImageType, daily: bool
    ):
        filters = self._get_search_filters(
            release=release, arch=arch, image_type=image_type, daily=daily
        )
        owner = self._get_owner(image_type=image_type)

        images = self.client.describe_images(
            Owners=[owner],
            Filters=filters,
        )

        if not images.get("Images"):
            raise ValueError(
                "Could not find {} image for {} release".format(
                    image_type.value, release
                )
            )

        return sorted(images["Images"], key=lambda x: x["CreationDate"])[-1]

    def daily_image(
        self,
        release: str,
        *,
        arch: str = "x86_64",
        image_type: ImageType = ImageType.GENERIC,
        **kwargs,
    ):
        """Find the id of the latest daily image for a particular release.

        Args:
            release: string, Ubuntu release to look for
            arch: string, architecture to use

        Returns:
            string, id of latest image

        """
        self._log.debug("finding daily Ubuntu image for %s", release)
        image = self._find_latest_image(
            release=release, arch=arch, image_type=image_type, daily=True
        )
        return image["ImageId"]

    def _find_image_serial(
        self, image_id, image_type: ImageType = ImageType.GENERIC
    ):
        owner = self._get_owner(image_type=image_type)
        filters = [
            {
                "Name": "image-id",
                "Values": (image_id,),
            }
        ]

        images = self.client.describe_images(
            Owners=[owner],
            Filters=filters,
        )

        if not images.get("Images"):
            raise ImageNotFoundError(image_id)

        image_name = images["Images"][0].get("Name", "")
        serial_regex = r"ubuntu/.*/.*/.*-(?P<serial>\d+(\.\d+)?)$"
        serial_match = re.match(serial_regex, image_name)

        if not serial_match:
            raise ImageNotFoundError(resource_id=image_id)

        return serial_match.groupdict().get("serial")

    def image_serial(
        self, image_id, image_type: ImageType = ImageType.GENERIC
    ):
        """Find the image serial of a given EC2 image ID.

        Args:
            image_id: string, Ubuntu image id

        Returns:
            string, serial of latest image

        """
        self._log.debug(
            "finding image serial for EC2 Ubuntu image %s", image_id
        )
        return self._find_image_serial(image_id, image_type)

    def delete_image(self, image_id, **kwargs):
        """Delete an image.

        Args:
            image_id: string, id of the image to delete
        """
        image = self.resource.Image(image_id)
        snapshot_id = image.block_device_mappings[0]["Ebs"]["SnapshotId"]

        self._log.debug("removing custom ami %s", image_id)
        self.client.deregister_image(ImageId=image_id)

        self._log.debug("removing custom snapshot %s", snapshot_id)
        self.client.delete_snapshot(SnapshotId=snapshot_id)

    def delete_key(self, name):
        """Delete an uploaded key.

        Args:
            name: The key name to delete.
        """
        self._log.debug("deleting SSH key %s", name)
        self.client.delete_key_pair(KeyName=name)

    def get_instance(self, instance_id):
        """Get an instance by id.

        Args:
            instance_id:

        Returns:
            An instance object to use to manipulate the instance further.

        """
        instance = self.resource.Instance(instance_id)
        return EC2Instance(self.key_pair, self.client, instance)

    def launch(
        self,
        image_id,
        instance_type="t3.micro",  # Using nitro instance for IPv6
        user_data=None,
        wait=True,
        vpc=None,
        **kwargs,
    ):
        """Launch instance on EC2.

        Args:
            image_id: string, AMI ID to use default: latest Ubuntu LTS
            instance_type: string, instance type to launch
            user_data: string, user-data to pass to instance
            wait: boolean, wait for instance to come up
            vpc: optional vpc object to create instance under
            kwargs: other named arguments to add to instance JSON

        Returns:
            EC2 Instance object
        Raises: ValueError on invalid image_id
        """
        if not image_id:
            raise ValueError(
                f"{self._type} launch requires image_id param."
                f" Found: {image_id}"
            )
        args = {
            "ImageId": image_id,
            "InstanceType": instance_type,
            "KeyName": self.key_pair.name,
            "MaxCount": 1,
            "MinCount": 1,
            "TagSpecifications": [
                {
                    "ResourceType": "instance",
                    "Tags": [{"Key": "Name", "Value": self.tag}],
                }
            ],
        }

        if user_data:
            args["UserData"] = user_data

        for key, value in kwargs.items():
            args[key] = value

        if vpc:
            try:
                [subnet_id] = [s.id for s in vpc.vpc.subnets.all()]
            except ValueError as e:
                raise PycloudlibError(
                    "Too many subnets in vpc {}. pycloudlib does not support"
                    " launching into VPCs with multiple subnets".format(vpc.id)
                ) from e
            args["SubnetId"] = subnet_id
            args["SecurityGroupIds"] = [
                sg.id for sg in vpc.vpc.security_groups.all()
            ]

        self._log.debug("launching instance")
        instances = self.resource.create_instances(**args)
        instance = EC2Instance(self.key_pair, self.client, instances[0])

        if wait:
            instance.wait()

        return instance

    def list_keys(self):
        """List all ssh key pair names loaded on this EC2 region."""
        keypair_names = []
        for keypair in self.client.describe_key_pairs()["KeyPairs"]:
            keypair_names.append(keypair["KeyName"])
        return keypair_names

    def snapshot(self, instance, clean=True):
        """Snapshot an instance and generate an image from it.

        Args:
            instance: Instance to snapshot
            clean: run instance clean method before taking snapshot

        Returns:
            An image id

        """
        if clean:
            instance.clean()

        instance.shutdown(wait=True)

        self._log.debug("creating custom ami from instance %s", instance.id)

        response = self.client.create_image(
            Name="%s-%s" % (self.tag, instance.image_id),
            InstanceId=instance.id,
        )
        image_ami_edited = response["ImageId"]
        image = self.resource.Image(image_ami_edited)

        self._wait_for_snapshot(image)
        _tag_resource(image, self.tag)

        instance.start(wait=True)

        return image.id

    def upload_key(self, public_key_path, private_key_path=None, name=None):
        """Use an existing already uploaded key.

        Args:
            public_key_path: path to the public key to upload
            private_key_path: path to the private key to upload
            name: name to reference key by
        """
        self._log.debug("uploading SSH key %s", name)
        self.client.import_key_pair(
            KeyName=name, PublicKeyMaterial=self.key_pair.public_key_content
        )
        self.use_key(public_key_path, private_key_path, name)

    def use_key(self, public_key_path, private_key_path=None, name=None):
        """Use an existing already uploaded key.

        Args:
            public_key_path: path to the public key to upload
            private_key_path: path to the private key to upload
            name: name to reference key by
        """
        if not name:
            name = self.tag
        super().use_key(public_key_path, private_key_path, name)

    def _wait_for_snapshot(self, image):
        """Wait for snapshot image to be created.

        Args:
            image: image boto3 object to wait to be available
        """
        image.wait_until_exists()
        waiter = self.client.get_waiter("image_available")
        waiter.wait(ImageIds=[image.id])
        image.reload()



# This file is part of pycloudlib. See LICENSE file for license information.
"""Base class for all other clouds to provide consistent set of functions."""

import enum
import getpass
import io
import logging
import os
from abc import ABC, abstractmethod
from typing import Any, Optional, Sequence

import paramiko

from pycloudlib.config import ConfigFile, parse_config
from pycloudlib.instance import BaseInstance
from pycloudlib.key import KeyPair
from pycloudlib.util import get_timestamped_tag, validate_tag

_RequiredValues = Optional[Sequence[Optional[Any]]]


@enum.unique
class ImageType(enum.Enum):
    """Allowed image types when launching cloud images."""

    GENERIC = "generic"
    PRO = "Pro"
    PRO_FIPS = "Pro FIPS"


class BaseCloud(ABC):
    """Base Cloud Class."""

    _type = "base"

    def __init__(
        self,
        tag: str,
        timestamp_suffix: bool = True,
        config_file: Optional[ConfigFile] = None,
        required_values: _RequiredValues = None,
    ):
        """Initialize base cloud class.

        Args:
            tag: string used to name and tag resources with
            timestamp_suffix: Append a timestamped suffix to the tag string.
            config_file: path to pycloudlib configuration file
        """
        self._log = logging.getLogger(
            "{}.{}".format(__name__, self.__class__.__name__)
        )
        self._check_and_set_config(config_file, required_values)

        user = getpass.getuser()
        self.key_pair = KeyPair(
            public_key_path=os.path.expandvars(
                self.config.get("public_key_path", f"~{user}/.ssh/id_rsa.pub")
            ),
            private_key_path=os.path.expandvars(
                self.config.get("private_key_path", "")
            ),
            name=self.config.get("key_name", user),
        )
        if timestamp_suffix:
            self.tag = validate_tag(get_timestamped_tag(tag))
        else:
            self.tag = validate_tag(tag)

    @abstractmethod
    def delete_image(self, image_id, **kwargs):
        """Delete an image.

        Args:
            image_id: string, id of the image to delete
            **kwargs: dictionary of other arguments to pass to delete_image
        """
        raise NotImplementedError

    @abstractmethod
    def released_image(self, release, **kwargs):
        """ID of the latest released image for a particular release.

        Args:
            release: The release to look for

        Returns:
            A single string with the latest released image ID for the
            specified release.

        """
        raise NotImplementedError

    @abstractmethod
    def daily_image(self, release: str, **kwargs):
        """ID of the latest daily image for a particular release.

        Args:
            release: The release to look for

        Returns:
            A single string with the latest daily image ID for the
            specified release.

        """
        raise NotImplementedError

    @abstractmethod
    def image_serial(self, image_id):
        """Find the image serial of the latest daily image for a particular release.

        Args:
            image_id: string, Ubuntu image id

        Returns:
            string, serial of latest image

        """
        raise NotImplementedError

    @abstractmethod
    def get_instance(self, instance_id, **kwargs) -> BaseInstance:
        """Get an instance by id.

        Args:
            instance_id:

        Returns:
            An instance object to use to manipulate the instance further.

        """
        raise NotImplementedError

    @abstractmethod
    def launch(
        self,
        image_id: str,
        instance_type=None,
        user_data=None,
        wait: bool = True,
        **kwargs,
    ) -> BaseInstance:
        """Launch an instance.

        Args:
            image_id: string, image ID to use for the instance
            instance_type: string, type of instance to create
            user_data: used by cloud-init to run custom scripts/configuration
            wait: wait for instance to be live
            **kwargs: dictionary of other arguments to pass to launch

        Returns:
            An instance object to use to manipulate the instance further.

        """
        raise NotImplementedError

    @abstractmethod
    def snapshot(self, instance, clean=True, **kwargs):
        """Snapshot an instance and generate an image from it.

        Args:
            instance: Instance to snapshot
            clean: run instance clean method before taking snapshot

        Returns:
            An image id

        """
        raise NotImplementedError

    def list_keys(self):
        """List ssh key names present on the cloud for accessing instances.

        Returns:
           A list of strings of key pair names accessible to the cloud.

        """
        raise NotImplementedError

    def create_key_pair(self):
        """Create and set a ssh key pair for a cloud instance.

        Returns:
            A tuple containing the public and private key created
        """
        key = paramiko.RSAKey.generate(4096)
        priv_str = io.StringIO()

        pub_key = "{} {}".format(key.get_name(), key.get_base64())
        key.write_private_key(priv_str, password=None)

        return pub_key, priv_str.getvalue()

    def use_key(self, public_key_path, private_key_path=None, name=None):
        """Use an existing key.

        Args:
            public_key_path: path to the public key to upload
            private_key_path: path to the private key
            name: name to reference key by
        """
        self._log.debug("using SSH key from %s", public_key_path)
        self.key_pair = KeyPair(public_key_path, private_key_path, name)

    def _check_and_set_config(
        self,
        config_file: Optional[ConfigFile],
        required_values: _RequiredValues,
    ):
        """Set pycloudlib configuration.

        Checks if values required to launch a cloud instance are present.
        Values should be present in pycloudlib config file or passed to the
        cloud's constructor directly.

        Args:
            config_file: path to pycloudlib configuration file
            required_values: a list containing all the required values for
                the cloud that were passed to the cloud's constructor
        """
        # if all required values were passed to the cloud's constructor,
        # there is no need to parse the config file. If some (but not all)
        # of them were provided, config file is loaded and the values that
        # were passed in work as overrides
        if required_values and all(v is not None for v in required_values):
            self.config = {}
        else:
            self.config = parse_config(config_file)[self._type]



"""Deal with configuration file."""
import logging
import os
from io import StringIO
from pathlib import Path
from typing import Any, MutableMapping, Optional, Union

import toml

# Order matters here. Local should take precedence over global.
CONFIG_PATHS = [
    Path("~/.config/pycloudlib.toml").expanduser(),
    Path("/etc/pycloudlib.toml"),
]

ConfigFile = Union[Path, StringIO]
log = logging.getLogger(__name__)


class Config(dict):
    """Override dict to allow raising a more meaningful KeyError."""

    def __getitem__(self, key):
        """Provide more meaningful KeyError on access."""
        try:
            return super().__getitem__(key)
        except KeyError:
            raise KeyError(
                "{} must be defined in pycloudlib.toml to make this "
                "call".format(key)
            ) from None


def parse_config(
    config_file: Optional[ConfigFile] = None,
) -> MutableMapping[str, Any]:
    """Find the relevant TOML, load, and return it."""
    possible_configs = []
    if config_file:
        possible_configs.append(config_file)
    if os.environ.get("PYCLOUDLIB_CONFIG"):
        possible_configs.append(Path(os.environ["PYCLOUDLIB_CONFIG"]))
    possible_configs.extend(CONFIG_PATHS)
    for path in possible_configs:
        try:
            config = toml.load(path, _dict=Config)
            log.debug("Loaded configuration from %s", path)
            return config
        except FileNotFoundError:
            continue
        except toml.TomlDecodeError as e:
            raise ValueError(
                "Could not parse configuration file pointed to by "
                "{}".format(path)
            ) from e
    raise ValueError(
        "No configuration file found! Copy pycloudlib.toml.template to "
        "~/.config/pycloudlib.toml or /etc/pycloudlib.toml"
    )



# pylint: disable=E1101
# This file is part of pycloudlib. See LICENSE file for license information.
"""OCI Cloud type."""

import base64
import json
import os
import re
from typing import Optional, cast

import oci

from pycloudlib.cloud import BaseCloud
from pycloudlib.config import ConfigFile
from pycloudlib.errors import CloudSetupError, InstanceNotFoundError
from pycloudlib.oci.instance import OciInstance
from pycloudlib.oci.utils import get_subnet_id, wait_till_ready
from pycloudlib.util import UBUNTU_RELEASE_VERSION_MAP, subp


class OCI(BaseCloud):
    """OCI (Oracle) cloud class."""

    _type = "oci"

    def __init__(
        self,
        tag: str,
        timestamp_suffix: bool = True,
        config_file: Optional[ConfigFile] = None,
        *,
        availability_domain: Optional[str] = None,
        compartment_id: Optional[str] = None,
        config_path: Optional[str] = None,
        config_dict: Optional[str] = None,
    ):
        """
        Initialize the connection to OCI.

        OCI must be initialized on the CLI first:
        https://github.com/cloud-init/qa-scripts/blob/master/doc/launching-oracle.md

        Args:
            tag: Name of instance
            timestamp_suffix: bool set True to append a timestamp suffix to the
                tag
            config_file: path to pycloudlib configuration file
            compartment_id: A compartment found at
                https://console.us-phoenix-1.oraclecloud.com/a/identity/compartments
            availability_domain: One of the availability domains from:
                'oci iam availability-domain list'
            config_path: Path of OCI config file
            config_dict: A dictionary containing the OCI config.
                Overrides the values from config_path
        """
        super().__init__(
            tag,
            timestamp_suffix,
            config_file,
            required_values=[availability_domain, compartment_id],
        )

        self.availability_domain = (
            availability_domain or self.config["availability_domain"]
        )

        compartment_id = compartment_id or self.config.get("compartment_id")
        if not compartment_id:
            command = ["oci", "iam", "compartment", "get"]
            exception_text = (
                "Could not obtain OCI compartment id. Has the CLI client been "
                "setup?\nCommand attempted: '{}'".format(" ".join(command))
            )
            try:
                result = subp(command, rcs=())
            except FileNotFoundError as e:
                raise CloudSetupError(exception_text) from e
            if not result.ok:
                exception_text += "\nstdout: {}\nstderr: {}".format(
                    result.stdout, result.stderr
                )
                raise CloudSetupError(exception_text)
            compartment_id = cast(str, json.loads(result.stdout)["data"]["id"])
        self.compartment_id = compartment_id

        if config_dict:
            try:
                oci.config.validate_config(config_dict)  # type: ignore
                self.oci_config = config_dict
            except oci.exceptions.InvalidConfig as e:  # type: ignore
                raise ValueError(
                    "Config dict is invalid. Pass a valid config dict. "
                    "{}".format(e)
                ) from e

        else:
            config_path = (
                config_path
                or self.config.get("config_path")
                or "~/.oci/config"
            )
            if not os.path.isfile(os.path.expanduser(config_path)):
                raise ValueError(
                    "{} is not a valid config file. Pass a valid config "
                    "file.".format(config_path)
                )
            self.oci_config = oci.config.from_file(config_path)  # type: ignore

        self._log.debug("Logging into OCI")
        self.compute_client = oci.core.ComputeClient(self.oci_config)  # type: ignore # noqa: E501
        self.network_client = oci.core.VirtualNetworkClient(self.oci_config)  # type: ignore # noqa: E501

    def delete_image(self, image_id, **kwargs):
        """Delete an image.

        Args:
            image_id: string, id of the image to delete
        """
        self.compute_client.delete_image(image_id, **kwargs)

    def released_image(self, release, operating_system="Canonical Ubuntu"):
        """Get the released image.

        OCI just has periodic builds, so "released" and "daily" don't
        really make sense here. Just call the same code for both

        Args:
            release: string, Ubuntu release to look for
            operating_system: string, operating system to use
        Returns:
            string, id of latest image

        """
        return self.daily_image(release, operating_system)

    def daily_image(
        self,
        release: str,
        operating_system: str = "Canonical Ubuntu",
        **kwargs,
    ):
        """Get the daily image.

        OCI just has periodic builds, so "released" and "daily" don't
        really make sense here. Just call the same code for both.

        Should be equivalent to the cli call:
        oci compute image list \
          --operating-system="Canonical Ubuntu" \
          --operating-system-version="<xx.xx>" \
          --sort-by='TIMECREATED' \
          --sort-order='DESC'

        Args:
            release: string, Ubuntu release to look for
            operating_system: string, Operating system to use
            **kwargs: dictionary of other arguments to pass to list_images

        Returns:
            string, id of latest image

        """
        if operating_system == "Canonical Ubuntu":
            if not re.match(r"^\d{2}\.\d{2}$", release):  # 18.04, 20.04, etc
                try:
                    release = UBUNTU_RELEASE_VERSION_MAP[release]
                except KeyError as e:
                    raise ValueError("Invalid release") from e

        # OCI likes to keep a few of each image around, so sort by
        # timestamp descending and grab the first (most recent) one
        image_response = self.compute_client.list_images(
            self.compartment_id,
            operating_system=operating_system,
            operating_system_version=release,
            sort_by="TIMECREATED",
            sort_order="DESC",
            **kwargs,
        )
        matching_image = [
            i
            for i in image_response.data
            if "aarch64" not in i.display_name and "GPU" not in i.display_name
        ]
        image_id = matching_image[0].id
        return image_id

    def image_serial(self, image_id):
        """Find the image serial of the latest daily image for a particular release.

        Args:
            image_id: string, Ubuntu image id

        Returns:
            string, serial of latest image

        """
        raise NotImplementedError

    def get_instance(self, instance_id, **kwargs) -> OciInstance:
        """Get an instance by id.

        Args:
            instance_id: ocid of the instance
            **kwargs: dictionary of other arguments to pass to get_instance
        Returns:
            An instance object to use to manipulate the instance further.
        """
        # verifies that instance id exists in oracle
        try:
            self.compute_client.get_instance(instance_id, **kwargs)
        except oci.exceptions.ServiceError as e:  # type: ignore
            raise InstanceNotFoundError(resource_id=instance_id) from e

        return OciInstance(
            key_pair=self.key_pair,
            instance_id=instance_id,
            compartment_id=self.compartment_id,
            availability_domain=self.availability_domain,
            oci_config=self.oci_config,
        )

    def launch(
        self,
        image_id,
        instance_type="VM.Standard2.1",
        user_data=None,
        wait=True,
        *,
        retry_strategy=None,
        **kwargs,
    ) -> OciInstance:
        """Launch an instance.

        Args:
            image_id: string, image ID to use for the instance
            instance_type: string, type of instance to create.
                https://docs.cloud.oracle.com/en-us/iaas/Content/Compute/References/computeshapes.htm
            user_data: used by Cloud-Init to run custom scripts or
                provide custom Cloud-Init configuration
            wait: wait for instance to be live
            retry_strategy: a retry strategy from oci.retry module
                to apply for this operation
            **kwargs: dictionary of other arguments to pass as
                LaunchInstanceDetails

        Returns:
            An instance object to use to manipulate the instance further.
        Raises: ValueError on invalid image_id
        """
        if not image_id:
            raise ValueError(
                f"{self._type} launch requires image_id param."
                f" Found: {image_id}"
            )
        subnet_id = get_subnet_id(
            self.network_client, self.compartment_id, self.availability_domain
        )
        metadata = {
            "ssh_authorized_keys": self.key_pair.public_key_content,
        }
        if user_data:
            metadata["user_data"] = base64.b64encode(
                user_data.encode("utf8")
            ).decode("ascii")

        instance_details = oci.core.models.LaunchInstanceDetails(  # type: ignore[attr-defined] # noqa: E501
            display_name=self.tag,
            availability_domain=self.availability_domain,
            compartment_id=self.compartment_id,
            shape=instance_type,
            subnet_id=subnet_id,
            image_id=image_id,
            metadata=metadata,
            **kwargs,
        )

        instance_data = self.compute_client.launch_instance(
            instance_details, retry_strategy=retry_strategy
        ).data
        instance = self.get_instance(
            instance_data.id, retry_strategy=retry_strategy
        )
        if wait:
            wait_till_ready(
                func=self.compute_client.get_instance,
                current_data=instance_data,
                desired_state="RUNNING",
            )
            instance.wait()
        return instance

    def snapshot(self, instance, clean=True, name=None):
        """Snapshot an instance and generate an image from it.

        Args:
            instance: Instance to snapshot
            clean: run instance clean method before taking snapshot
            name: (Optional) Name of created image
        Returns:
            An image object
        """
        if clean:
            instance.clean()
        image_details = {
            "compartment_id": self.compartment_id,
            "instance_id": instance.instance_id,
        }
        if name:
            image_details["display_name"] = name
        image_data = self.compute_client.create_image(
            oci.core.models.CreateImageDetails(**image_details)
        ).data
        image_data = wait_till_ready(
            func=self.compute_client.get_image,
            current_data=image_data,
            desired_state="AVAILABLE",
        )

        return image_data.id



# This file is part of pycloudlib. See LICENSE file for license information.
"""LXD Cloud type."""
import warnings
from abc import ABC
from itertools import count

import yaml

from pycloudlib.cloud import BaseCloud
from pycloudlib.constants import LOCAL_UBUNTU_ARCH
from pycloudlib.lxd import _images
from pycloudlib.lxd.defaults import base_vm_profiles
from pycloudlib.lxd.instance import LXDInstance, LXDVirtualMachineInstance
from pycloudlib.util import subp


class _BaseLXD(BaseCloud, ABC):
    """LXD Base Cloud Class."""

    _type = "lxd"
    _daily_remote = "ubuntu-daily"
    _releases_remote = "ubuntu"
    _lxd_instance_cls = LXDInstance
    _instance_counter = count()
    _is_container: bool

    def clone(self, base, new_instance_name):
        """Create copy of an existing instance or snapshot.

        Uses the `lxc copy` command to create a copy of an existing
        instance or a snapshot. To clone a snapshot then the base
        is `instance_name/snapshot_name` otherwise if base is only
        an existing instance it will clone an instance.

        Args:
            base: base instance or instance/snapshot
            new_instance_name: name of new instance

        Returns:
            The created LXD instance object

        """
        self._log.debug("cloning %s to %s", base, new_instance_name)
        subp(["lxc", "copy", base, new_instance_name])
        return LXDInstance(new_instance_name)

    def create_profile(self, profile_name, profile_config, force=False):
        """Create a lxd profile.

        Create a lxd profile and populate it with the given
        profile config. If the profile already exists, we will
        not recreate it, unless the force parameter is set to True.

        Args:
            profile_name: Name of the profile to be created
            profile_config: Config to be added to the new profile
            force: Force the profile creation if it already exists
        """
        profile_yaml = subp(["lxc", "profile", "list", "--format", "yaml"])
        profile_list = [
            profile["name"] for profile in yaml.safe_load(profile_yaml)
        ]

        if profile_name in profile_list and not force:
            msg = "The profile named {} already exists".format(profile_name)
            self._log.debug(msg)
            print(msg)
            return

        if force:
            self._log.debug("Deleting current profile %s ...", profile_name)
            subp(["lxc", "profile", "delete", profile_name])

        self._log.debug("Creating profile %s ...", profile_name)
        subp(["lxc", "profile", "create", profile_name])
        subp(["lxc", "profile", "edit", profile_name], data=profile_config)

    def delete_instance(self, instance_name, wait=True):
        """Delete an instance.

        Args:
            instance_name: instance name to delete
            wait: wait for delete to complete
        """
        self._log.debug("deleting %s", instance_name)
        inst = self.get_instance(instance_name)
        inst.delete(wait)

    def get_instance(self, instance_id):
        """Get an existing instance.

        Args:
            instance_id: instance name to get

        Returns:
            The existing instance as a LXD instance object

        """
        return self._lxd_instance_cls(instance_id, key_pair=self.key_pair)

    def _normalize_image_id(self, image_id: str) -> str:
        if ":" not in image_id:
            return self._daily_remote + ":" + image_id
        return image_id

    # pylint: disable=R0914,R0912,R0915
    def _prepare_command(
        self,
        name,
        image_id,
        ephemeral=False,
        network=None,
        storage=None,
        inst_type=None,
        profile_list=None,
        user_data=None,
        config_dict=None,
    ):
        """Build a the command to be used to launch the LXD instance.

        Args:
            name: string, what to call the instance
            image_id: string, [<remote>:]<image identifier>, the image to
                      launch
            ephemeral: boolean, ephemeral, otherwise persistent
            network: string, optional, network name to use
            storage: string, optional, storage name to use
            inst_type: string, optional, type to use
            profile_list: list, optional, profile(s) to use
            user_data: used by cloud-init to run custom scripts/configuration
            config_dict: dict, optional, configuration values to pass

        Returns:
            A list of string representing the command to be run to
            launch the LXD instance.
        """
        profile_list = profile_list if profile_list else []
        config_dict = config_dict if config_dict else {}

        self._log.debug("Full image ID to launch: '%s'", image_id)
        cmd = ["lxc", "init", image_id]

        if name:
            cmd.append(name)

        if self.key_pair:
            metadata = "public-keys: {}".format(
                self.key_pair.public_key_content
            )
            config_dict["user.meta-data"] = metadata

        if ephemeral:
            cmd.append("--ephemeral")

        if network:
            cmd.append("--network")
            cmd.append(network)

        if storage:
            cmd.append("--storage")
            cmd.append(storage)

        if inst_type:
            cmd.append("--type")
            cmd.append(inst_type)

        for profile in profile_list:
            cmd.append("--profile")
            cmd.append(profile)

        for key, value in config_dict.items():
            cmd.append("--config")
            cmd.append("%s=%s" % (key, value))

        if user_data:
            if "user.user-data" in config_dict:
                raise ValueError(
                    "User data cannot be defined in config_dict and also"
                    "passed through user_data. Pick one"
                )
            cmd.append("--config")
            cmd.append("user.user-data=%s" % user_data)

        return cmd

    def init(
        self,
        name,
        image_id,
        ephemeral=False,
        network=None,
        storage=None,
        inst_type=None,
        profile_list=None,
        user_data=None,
        config_dict=None,
        execute_via_ssh=True,
    ):
        """Init a container.

        This will initialize a container, but not launch or start it.
        If no remote is specified pycloudlib default to daily images.

        Args:
            name: string, what to call the instance
            image_id: string, [<remote>:]<image identifier>, the image to
                      launch
            ephemeral: boolean, ephemeral, otherwise persistent
            network: string, optional, network name to use
            storage: string, optional, storage name to use
            inst_type: string, optional, type to use
            profile_list: list, optional, profile(s) to use
            user_data: used by cloud-init to run custom scripts/configuration
            config_dict: dict, optional, configuration values to pass
            execute_via_ssh: bool, optional, execute commands on the instance
                             via SSH if True (the default)

        Returns:
            The created LXD instance object

        """
        image_id = self._normalize_image_id(image_id)
        series = _images.find_release(image_id)

        cmd = self._prepare_command(
            name=name,
            image_id=image_id,
            ephemeral=ephemeral,
            network=network,
            storage=storage,
            inst_type=inst_type,
            profile_list=profile_list,
            user_data=user_data,
            config_dict=config_dict,
        )

        print(cmd)
        result = subp(cmd)

        if not name:
            name = result.split("Instance name is: ")[1]

        self._log.debug("Created %s", name)
        return self._lxd_instance_cls(
            name=name,
            key_pair=self.key_pair,
            execute_via_ssh=execute_via_ssh,
            series=series,
            ephemeral=ephemeral,
        )

    def launch(
        self,
        image_id,
        instance_type=None,
        user_data=None,
        wait=True,
        name=None,
        ephemeral=False,
        network=None,
        storage=None,
        profile_list=None,
        config_dict=None,
        execute_via_ssh=True,
        **kwargs,
    ):
        """Set up and launch a container.

        This will init and start a container with the provided settings.
        If no remote is specified pycloudlib defaults to daily images.

        Args:
            image_id: string, [<remote>:]<image>, the image to launch
            instance_type: string, type to use
            user_data: used by cloud-init to run custom scripts/configuration
            wait: boolean, wait for instance to start
            name: string, what to call the instance
            ephemeral: boolean, ephemeral, otherwise persistent
            network: string, network name to use
            storage: string, storage name to use
            profile_list: list, profile(s) to use
            config_dict: dict, configuration values to pass
            execute_via_ssh: bool, optional, execute commands on the instance
                             via SSH if True (the default)

        Returns:
            The created LXD instance object
        Raises: ValueError on missing image_id
        """
        if not image_id:
            raise ValueError(
                f"{self._type} launch requires image_id param."
                f" Found: {image_id}"
            )
        instance = self.init(
            name=name or f"{self.tag}-{next(self._instance_counter)}",
            image_id=image_id,
            ephemeral=ephemeral,
            network=network,
            storage=storage,
            inst_type=instance_type,
            profile_list=profile_list,
            user_data=user_data,
            config_dict=config_dict,
            execute_via_ssh=execute_via_ssh,
        )
        instance.start(wait)

        return instance

    def released_image(self, release, arch=LOCAL_UBUNTU_ARCH):
        """Find the LXD fingerprint of the latest released image.

        Args:
            release: string, Ubuntu release to look for
            arch: string, architecture to use

        Returns:
            string, LXD fingerprint of latest image

        """
        self._log.debug("finding released Ubuntu image for %s", release)
        return _images.find_last_fingerprint(
            daily=False,
            release=release,
            arch=arch,
            is_container=self._is_container,
        )

    def daily_image(
        self, release: str, arch: str = LOCAL_UBUNTU_ARCH, **kwargs
    ):
        """Find the LXD fingerprint of the latest daily image.

        Args:
            release: string, Ubuntu release to look for
            arch: string, architecture to use

        Returns:
            string, LXD fingerprint of latest image

        """
        self._log.debug("finding daily Ubuntu image for %s", release)
        return _images.find_last_fingerprint(
            daily=True,
            release=release,
            arch=arch,
            is_container=self._is_container,
        )

    def image_serial(self, image_id):
        """Find the image serial of a given LXD image.

        Args:
            image_id: string, LXD image fingerprint

        Returns:
            string, serial of latest image

        """
        self._log.debug(
            "finding image serial for LXD Ubuntu image %s", image_id
        )
        return _images.find_image_serial(image_id)

    def delete_image(self, image_id, **kwargs):
        """Delete the image.

        Args:
            image_id: string, LXD image fingerprint
        """
        self._log.debug("Deleting image: '%s'", image_id)

        subp(["lxc", "image", "delete", image_id])
        self._log.debug("Deleted %s", image_id)

    def snapshot(self, instance, clean=True, name=None):
        """Take a snapshot of the passed in instance for use as image.

        :param instance: The instance to create an image from
        :type instance: LXDInstance
        :param clean: Whether to call cloud-init clean before creation
        :param wait: Whether to wait until before image is created
            before returning
        :param name: Name of the new image
        :param stateful: Whether to use an LXD stateful snapshot
        """
        if clean:
            instance.clean()

        return instance.snapshot(name)


class LXDContainer(_BaseLXD):
    """LXD Containers Cloud Class."""

    def __init__(self, *args, **kwargs):
        """Run LXDContainer constructor."""
        super().__init__(*args, **kwargs)
        self._is_container = True


class LXD(LXDContainer):
    """Old LXD Container Cloud Class (Kept for compatibility issues)."""

    def __init__(self, *args, **kwargs):
        """Run LXDContainer constructor."""
        warnings.warn("LXD class is deprecated; use LXDContainer instead.")
        super().__init__(*args, **kwargs)


class LXDVirtualMachine(_BaseLXD):
    """LXD Virtual Machine Cloud Class."""

    _lxd_instance_cls = LXDVirtualMachineInstance

    def __init__(self, *args, **kwargs):
        """Run LXDVirtualMachine constructor."""
        super().__init__(*args, **kwargs)
        self._is_container = False

    def build_necessary_profiles(self, image_id):
        """Build necessary profiles to launch the LXD instance.

        Args:
            image_id: string, [<remote>:]<release>, the image to build profiles
                      for

        Returns:
            A list containing the profiles created
        """
        image_id = self._normalize_image_id(image_id)
        base_release = _images.find_release(image_id)
        if base_release not in ["xenial", "bionic"]:
            base_release = "default"
        profile_name = f"pycloudlib-vm-{base_release}"

        self.create_profile(
            profile_name=profile_name,
            profile_config=base_vm_profiles[base_release],
        )

        return [profile_name]

    def _prepare_command(
        self,
        name,
        image_id,
        ephemeral=False,
        network=None,
        storage=None,
        inst_type=None,
        profile_list=None,
        user_data=None,
        config_dict=None,
    ):
        """Build a the command to be used to launch the LXD instance.

        Args:
            name: string, what to call the instance
            image_id: string, [<remote>:]<image identifier>, the image to
                      launch
            ephemeral: boolean, ephemeral, otherwise persistent
            network: string, optional, network name to use
            storage: string, optional, storage name to use
            inst_type: string, optional, type to use
            profile_list: list, optional, profile(s) to use
            user_data: used by cloud-init to run custom scripts/configuration
            config_dict: dict, optional, configuration values to pass

        Returns:
            A list of string representing the command to be run to
            launch the LXD instance.
        """
        if not profile_list:
            profile_list = self.build_necessary_profiles(image_id)

        cmd = super()._prepare_command(
            name=name,
            image_id=image_id,
            ephemeral=ephemeral,
            network=network,
            storage=storage,
            inst_type=inst_type,
            profile_list=profile_list,
            user_data=user_data,
            config_dict=config_dict,
        )

        cmd.append("--vm")

        return cmd



# pylint: disable=E1101
# This file is part of pycloudlib. See LICENSE file for license information.
"""OCI instance."""

from time import sleep

import oci

from pycloudlib.errors import PycloudlibError
from pycloudlib.instance import BaseInstance
from pycloudlib.oci.utils import get_subnet_id, wait_till_ready


class OciInstance(BaseInstance):
    """OCI backed instance."""

    _type = "oci"

    def __init__(
        self,
        key_pair,
        instance_id,
        compartment_id,
        availability_domain,
        oci_config=None,
    ):
        """Set up the instance.

        Args:
            key_pair: A KeyPair for SSH interactions
            instance_id: The instance id representing the cloud instance
            compartment_id: A compartment found at
                https://console.us-phoenix-1.oraclecloud.com/a/identity/compartments
            availability_domain: One of the availability domains from:
                'oci iam availability-domain list'
            oci_config: OCI configuration dictionary

        """
        super().__init__(key_pair)
        self.instance_id = instance_id
        self.compartment_id = compartment_id
        self.availability_domain = availability_domain
        self._ip = None

        if oci_config is None:
            oci_config = oci.config.from_file("~/.oci/config")
        self.compute_client = oci.core.ComputeClient(oci_config)
        self.network_client = oci.core.VirtualNetworkClient(oci_config)

    def __repr__(self):
        """Create string representation of class."""
        return "{}(instance_id={}, compartment_id={})".format(
            self.__class__.__name__,
            self.instance_id,
            self.compartment_id,
        )

    @property
    def name(self):
        """Return the instance name."""
        return self.instance_id

    @property
    def ip(self):
        """Return IP address of instance."""
        if not self._ip:
            vnic_attachment = self.compute_client.list_vnic_attachments(
                compartment_id=self.compartment_id,
                instance_id=self.instance_data.id,
            ).data[0]

            self._ip = self.network_client.get_vnic(
                vnic_attachment.vnic_id
            ).data.public_ip
        return self._ip

    @property
    def instance_data(self):
        """Return JSON formatted details from OCI about this instance."""
        return self.compute_client.get_instance(self.instance_id).data

    def console_log(self):
        """Not currently implemented."""
        # This is possible, but we need to capture console history first
        # self.compute_client.capture_console_history(...)
        # self.compute_client.get_console_history_content(...)
        raise NotImplementedError

    def delete(self, wait=True):
        """Delete the instance.

        Args:
            wait: wait for instance to be deleted
        """
        self.compute_client.terminate_instance(self.instance_data.id)
        if wait:
            self.wait_for_delete()

    def _do_restart(self, **kwargs):
        """Restart the instance."""
        last_exception = None
        for _ in range(30):
            try:
                self.compute_client.instance_action(
                    self.instance_data.id, "RESET"
                )
                return
            except oci.exceptions.ServiceError as e:
                last_exception = e
                if last_exception.status != 409:
                    raise
                self._log.debug(
                    "Received 409 attempting to RESET instance. Retrying"
                )
                sleep(0.5)
        if last_exception:
            raise last_exception

    def shutdown(self, wait=True, **kwargs):
        """Shutdown the instance.

        Args:
            wait: wait for the instance to shutdown
        """
        self._sync_filesystem()
        self.compute_client.instance_action(self.instance_data.id, "STOP")
        if wait:
            self.wait_for_stop()

    def start(self, wait=True):
        """Start the instance.

        Args:
            wait: wait for the instance to start.
        """
        self.compute_client.instance_action(self.instance_data.id, "START")
        if wait:
            self.wait()

    def _wait_for_instance_start(self):
        """Wait for instance to be up."""
        wait_till_ready(
            func=self.compute_client.get_instance,
            current_data=self.instance_data,
            desired_state="RUNNING",
        )

    def wait_for_delete(self):
        """Wait for instance to be deleted."""
        wait_till_ready(
            func=self.compute_client.get_instance,
            current_data=self.instance_data,
            desired_state="TERMINATED",
        )

    def wait_for_stop(self):
        """Wait for instance stop."""
        wait_till_ready(
            func=self.compute_client.get_instance,
            current_data=self.instance_data,
            desired_state="STOPPED",
        )

    def add_network_interface(self) -> str:
        """Add network interface to running instance.

        Creates a nic and attaches it to the instance. This is effectively a
        hot-add of a network device. Returns the IP address of the added
        network interface as a string.

        Note: It assumes the associated compartment has at least one subnet and
        creates the vnic in the first encountered subnet.
        """
        subnet_id = get_subnet_id(
            self.network_client, self.compartment_id, self.availability_domain
        )
        create_vnic_details = oci.core.models.CreateVnicDetails(  # type: ignore # noqa: E501
            subnet_id=subnet_id,
        )
        attach_vnic_details = oci.core.models.AttachVnicDetails(  # type: ignore # noqa: E501
            create_vnic_details=create_vnic_details,
            instance_id=self.instance_id,
        )
        vnic_attachment_data = self.compute_client.attach_vnic(
            attach_vnic_details
        ).data
        vnic_attachment_data = wait_till_ready(
            func=self.compute_client.get_vnic_attachment,
            current_data=vnic_attachment_data,
            desired_state=vnic_attachment_data.LIFECYCLE_STATE_ATTACHED,
        )
        vnic_data = self.network_client.get_vnic(
            vnic_attachment_data.vnic_id
        ).data
        return vnic_data.private_ip

    def remove_network_interface(self, ip_address: str):
        """Remove network interface based on IP address.

        Find the NIC from the IP, detach from the instance.

        Note: In OCI, detaching triggers deletion.
        """
        vnic_attachments = oci.pagination.list_call_get_all_results_generator(  # type: ignore # noqa: E501
            self.compute_client.list_vnic_attachments,
            "record",
            self.compartment_id,
            instance_id=self.instance_id,
        )
        for vnic_attachment in vnic_attachments:
            vnic_data = self.network_client.get_vnic(
                vnic_attachment.vnic_id
            ).data
            if vnic_data.private_ip == ip_address:
                try:
                    self.compute_client.detach_vnic(vnic_attachment.id)
                except oci.exceptions.ServiceError:  # type: ignore
                    self._log.debug(
                        "Failed manually detaching and deleting network "
                        "interface. Interface should get destroyed on instance"
                        " cleanup."
                    )
                return
        raise PycloudlibError(
            f"Network interface with ip_address={ip_address} did not detach"
        )



# This file is part of pycloudlib. See LICENSE file for license information.
# pylint: disable=too-many-public-methods
"""Base class for all instances to provide consistent set of functions."""

import logging
from enum import Enum, auto
from functools import partial
from itertools import chain
from typing import TYPE_CHECKING, Callable, Optional

from ibm_cloud_sdk_core import ApiException, DetailedResponse
from ibm_vpc import VpcV1
from ibm_vpc.vpc_v1 import Instance as _Instance
from ibm_vpc.vpc_v1 import InstanceAction as _InstanceAction

from pycloudlib.ibm._util import get_first as _get_first
from pycloudlib.ibm._util import iter_resources as _iter_resources
from pycloudlib.ibm._util import wait_until as _wait_until
from pycloudlib.ibm.errors import IBMException
from pycloudlib.instance import BaseInstance

if TYPE_CHECKING:
    from typing_extensions import TypeAlias

    _Action: TypeAlias = _InstanceAction.TypeEnum
    _Status: TypeAlias = _Instance.StatusEnum
else:
    _Action = _InstanceAction.TypeEnum
    _Status = _Instance.StatusEnum

logger = logging.getLogger(__name__)
VpcV1Fn = Callable[..., DetailedResponse]


class _Subnet:
    def __init__(self, client: VpcV1, subnet: dict):
        self._client = client
        self._subnet = subnet

    @classmethod
    def create(
        cls,
        client: VpcV1,
        *,
        name: str,
        zone: str,
        resource_group_id: str,
        vpc_id: str,
    ):
        """Create Subnet."""
        subnet_proto = {
            "name": name,
            "resource_group": {"id": resource_group_id},
            "vpc": {"id": vpc_id},
            "total_ipv4_address_count": 256,
            "zone": {"name": zone},
        }
        subnet = client.create_subnet(subnet_proto).get_result()
        return cls(client, subnet)

    @classmethod
    def from_existing(cls, client: VpcV1, name: str, vpc_id: str) -> "_Subnet":
        """Instantiate Subnet by name."""
        subnet = _get_first(
            client.list_subnets,
            resource_name="subnets",
            filter_fn=(
                lambda subnet: subnet["vpc"]["id"] == vpc_id
                and subnet["name"] == name
            ),
        )
        if subnet is None:
            raise IBMException(f"Subnet not found: {name}")
        return cls(client, subnet)

    @classmethod
    def from_default(cls, client: VpcV1, zone: str, vpc_id: str) -> "_Subnet":
        """Find the default Subnet."""
        return cls.from_existing(client, f"{zone}-default-subnet", vpc_id)

    @classmethod
    def discover(cls, client: VpcV1, vpc_id: str) -> "_Subnet":
        """Discover a Subnet within a VPC."""
        subnet = _get_first(
            client.list_subnets,
            resource_name="subnets",
            filter_fn=(lambda subnet: subnet["vpc"]["id"] == vpc_id),
        )
        if subnet is None:
            raise IBMException(f"No subnet associated to vpc found: {vpc_id}")
        return cls(client, subnet)

    @property
    def id(self) -> str:
        """Subnet ID."""
        return self._subnet["id"]

    def _refresh(self) -> dict:
        self._subnet = self._client.get_subnet(self.id).get_result()
        return self._subnet

    def _wait_for_delete(self, sleep_seconds: int = 30):
        def _check_fn():
            try:
                self._refresh()
            except ApiException as e:
                if e.code == 404:
                    return True  # Instance deleted
                raise
            return False

        msg = (
            f"Subnet not terminated after {sleep_seconds} seconds. "
            "Check IBM VPC console."
        )

        _wait_until(
            _check_fn,
            timeout_seconds=sleep_seconds,
            timeout_msg_fn=lambda: msg,
        )

    def delete(self):
        """Delete Subnet."""
        self._client.delete_subnet(self.id)
        self._wait_for_delete()


class VPC:
    """Virtual Private Cloud class proxy for IBM VPC resource."""

    def __init__(
        self,
        key_pair,
        *,
        client: VpcV1,
        vpc: dict,
        resource_group_id: str,
        subnet: Optional[_Subnet] = None,
        **_kwargs,
    ):
        """Init a `VPC`."""
        self._key_pair = key_pair
        self._client = client
        self._vpc = vpc
        self._subnet = subnet
        self._resource_group_id = resource_group_id

    @classmethod
    def create(
        cls,
        *args,
        client: VpcV1,
        name: str,
        resource_group_id: str,
        zone: str,
        **kwargs,
    ) -> "VPC":
        """Create a VPC.

        Creates a Subnet and adds inbound rule to accept SSH connections to the
        default Security Group.
        """
        resource_group = {"id": resource_group_id}
        vpc = client.create_vpc(
            name=name, resource_group=resource_group
        ).get_result()

        # Allow SSH access
        default_sg = client.get_vpc_default_security_group(
            vpc["id"]
        ).get_result()
        rule_proto = {
            "direction": "inbound",
            "ip_version": "ipv4",
            "remote": {"cidr_block": "0.0.0.0/0"},
            "port_max": 22,
            "port_min": 22,
            "protocol": "tcp",
        }
        client.create_security_group_rule(
            default_sg["id"], rule_proto
        ).get_result()

        subnet = _Subnet.create(
            client,
            name=f"{name}-subnet",
            zone=zone,
            resource_group_id=resource_group_id,
            vpc_id=vpc["id"],
        )
        return cls(
            *args,
            client=client,
            vpc=vpc,
            resource_group_id=resource_group_id,
            subnet=subnet,
            **kwargs,
        )

    @classmethod
    def from_existing(
        cls,
        *args,
        client: VpcV1,
        name: str,
        resource_group_id: str,
        zone: str,
        **kwargs,
    ) -> "VPC":
        """Find a VPC by name.

        Try to discover a Subnet within it or create it if not found.
        """
        vpc = _get_first(
            client.list_vpcs,
            resource_name="vpcs",
            filter_fn=lambda vpc: vpc["name"] == name,
        )
        if vpc is None:
            raise IBMException(f"VPC not found: {name}")

        try:
            subnet = _Subnet.discover(client, vpc_id=vpc["id"])
        except IBMException:
            subnet = _Subnet.create(
                client,
                name=f"{name}-subnet",
                zone=zone,
                resource_group_id=resource_group_id,
                vpc_id=vpc["id"],
            )

        return cls(
            *args,
            client=client,
            vpc=vpc,
            resource_group_id=resource_group_id,
            subnet=subnet,
            **kwargs,
        )

    @classmethod
    def from_default(
        cls,
        *args,
        client: VpcV1,
        resource_group_id: str,
        region: str,
        zone: str,
        **kwargs,
    ) -> "VPC":
        """Find the `default` VPC and Subnet."""
        default_name = f"{region}-default-vpc"
        vpc = _get_first(
            client.list_vpcs,
            resource_name="vpcs",
            filter_fn=lambda vpc: vpc["name"] == default_name,
            resource_group_id=resource_group_id,
        )
        if vpc is None:
            raise IBMException(f"VPC not found: {default_name}")
        subnet = _Subnet.from_default(client, zone=zone, vpc_id=vpc["id"])
        return cls(
            *args,
            client=client,
            vpc=vpc,
            subnet=subnet,
            resource_group_id=resource_group_id,
            **kwargs,
        )

    @property
    def id(self) -> str:
        """VPC ID."""
        return self._vpc["id"]

    @property
    def name(self) -> str:
        """VPC name."""
        return self._vpc["name"]

    @property
    def subnet_id(self) -> str:
        """Subnet ID."""
        if self._subnet is None:
            raise IBMException("No subnet available")
        return self._subnet.id

    def delete(self) -> None:
        """Delete VPC.

        Note: This will delete all instances and subnets living in the VPC.
        """
        logger.info("Deleting VPC: %s", self.id)

        # Delete all instances of types contained in `_IBMInstanceType`
        instances_in_vpc = chain.from_iterable(
            map(
                lambda iit: _iter_resources(
                    partial(iit.list_instances, self._client),
                    resource_name="instances",
                    map_fn=lambda inst: IBMInstance.from_existing(
                        self._key_pair, client=self._client, instance=inst
                    ),
                    vpc_id=self.id,
                ),
                iter(_IBMInstanceType),
            )
        )

        for instance in instances_in_vpc:
            instance.delete(wait=True)

        if self._subnet is not None:
            self._subnet.delete()
        self._client.delete_vpc(self.id)


class _IBMInstanceType(Enum):
    """Abstracts the different instance types present in IBM VPC."""

    VSI = auto()
    BARE_METAL_SERVER = auto()

    @classmethod
    def from_instance_type(cls, instance_type: str) -> "_IBMInstanceType":
        """Translate from `instance_type` to `_IBMInstanceType`.

        Note: In IBM VPC terms, `instance_type`s are `profile`s.
        """
        if "metal" in instance_type:
            return cls.BARE_METAL_SERVER
        if "host" in instance_type:
            logger.warning(
                "%s instance_type looks like a Dedicated Host,"
                " which is not supported.",
                instance_type,
            )
        return cls.VSI

    @classmethod
    def from_raw_instance(cls, instance: dict) -> "_IBMInstanceType":
        """Determine `self` from a raw instance."""
        instance_type = instance["profile"]["name"]
        return cls.from_instance_type(instance_type)

    def create_instance(
        self, client: VpcV1, *args, **kwargs
    ) -> DetailedResponse:
        """Create instance."""
        if self == self.VSI:
            return client.create_instance(*args, **kwargs)
        if self == self.BARE_METAL_SERVER:
            return client.create_bare_metal_server(*args, **kwargs)
        raise NotImplementedError(f"Implement me for: {self}")

    def list_instances(self, client: VpcV1, **kwargs) -> DetailedResponse:
        """List instances."""
        if self == self.VSI:
            return client.list_instances(**kwargs)
        if self == self.BARE_METAL_SERVER:
            return client.list_bare_metal_servers(**kwargs)
        raise NotImplementedError(f"Implement me for: {self}")

    def delete_instance(
        self, client: VpcV1, *args, **kwargs
    ) -> DetailedResponse:
        """Delete instance."""
        if self == self.VSI:
            return client.delete_instance(*args, **kwargs)
        if self == self.BARE_METAL_SERVER:
            return client.delete_bare_metal_server(*args, **kwargs)
        raise NotImplementedError(f"Implement me for: {self}")

    def get_instance(
        self, client: VpcV1, instance_id: str, **kwargs
    ) -> DetailedResponse:
        """Get instance."""
        if self == self.VSI:
            return client.get_instance(instance_id, **kwargs)
        if self == self.BARE_METAL_SERVER:
            return client.get_bare_metal_server(instance_id, **kwargs)
        raise NotImplementedError(f"Implement me for: {self}")

    def execute_instance_action(
        self,
        client: VpcV1,
        *,
        id: str,  # pylint: disable=redefined-builtin
        action: _Action,
        force: Optional[bool] = False,
    ) -> DetailedResponse:
        """Execute instance action."""
        # Note: None of the these endpoints returns a query-able resource.
        # Thus, the only way to check if the action has been completed is
        # to directly retrieve the raw instance data.
        if self == self.VSI:
            return client.create_instance_action(id, action.value, force=force)
        if self == self.BARE_METAL_SERVER:
            if action == _Action.STOP:
                return client.stop_bare_metal_server(id)
            if action == _Action.START:
                return client.start_bare_metal_server(id)
            if action == _Action.REBOOT:
                return client.restart_bare_metal_server(id)
            raise NotImplementedError(f"Implement me for: {action}")
        raise NotImplementedError(f"Implement me for: {self}")

    def list_instance_network_interface_floating_ips(
        self, client: VpcV1, *args, **kwargs
    ) -> DetailedResponse:
        """List Floating IPs associated to a nic."""
        if self == self.VSI:
            return client.list_instance_network_interface_floating_ips(
                *args, **kwargs
            )
        if self == self.BARE_METAL_SERVER:
            return (
                client.list_bare_metal_server_network_interface_floating_ips(
                    *args, *kwargs
                )
            )
        raise NotImplementedError(f"Implement me for: {self}")

    def add_instance_network_interface_floating_ip(
        self, client: VpcV1, *, instance_id: str, **kwargs
    ) -> DetailedResponse:
        """Add Floating IP to an instance."""
        if self == self.VSI:
            return client.add_instance_network_interface_floating_ip(
                instance_id=instance_id, **kwargs
            )
        if self == self.BARE_METAL_SERVER:
            return client.add_bare_metal_server_network_interface_floating_ip(
                bare_metal_server_id=instance_id,
                **kwargs,
            )
        raise NotImplementedError(f"Implement me for: {self}")


class IBMInstance(BaseInstance):
    """Base instance object."""

    _type = "ibm"

    def __init__(
        self,
        key_pair,
        *,
        client: VpcV1,
        instance: dict,
        floating_ip: Optional[dict] = None,
    ):
        """Set up instance."""
        super().__init__(key_pair)

        self._client = client
        self._instance = instance
        self._floating_ip = floating_ip

        # mount methods that depend on `_IBMInstanceType`:
        self._ibm_instance_type = _IBMInstanceType.from_raw_instance(instance)
        self._delete_instance = partial(
            self._ibm_instance_type.delete_instance, self._client
        )
        self._get_instance = partial(
            self._ibm_instance_type.get_instance, self._client
        )
        self._execute_instance_action = partial(
            self._ibm_instance_type.execute_instance_action,
            self._client,
            id=self.id,
        )

    @classmethod
    def with_floating_ip(
        cls, *args, client: VpcV1, instance: dict, floating_ip: dict, **kwargs
    ) -> "IBMInstance":
        """Instantiate `self` from `instance` associated to `floating_ip`."""
        nic_id = instance["primary_network_interface"]["id"]

        ibm_instance_type = _IBMInstanceType.from_raw_instance(instance)
        ibm_instance_type.add_instance_network_interface_floating_ip(
            client,
            id=floating_ip["id"],
            instance_id=instance["id"],
            network_interface_id=nic_id,
        ).get_result()

        return cls(
            *args,
            client=client,
            instance=instance,
            floating_ip=floating_ip,
            **kwargs,
        )

    @classmethod
    def from_existing(
        cls, *args, client: VpcV1, instance: dict, **kwargs
    ) -> "IBMInstance":
        """Instantiate `self` from `instance`.

        If `floating_ip` is not given, it will try to discover an associated
        Floating Ip.
        """
        floating_ip = kwargs.pop(
            "floating_ip", None
        ) or cls._discover_floating_ip(client, instance)
        return cls(
            *args,
            client=client,
            instance=instance,
            floating_ip=floating_ip,
            **kwargs,
        )

    @classmethod
    def find_existing(
        cls, *args, client: VpcV1, instance_id: str, **kwargs
    ) -> "IBMInstance":
        """Find an instance by ID."""
        instance = _IBMInstanceType.VSI.get_instance(client, instance_id)
        if not instance:
            instance = _IBMInstanceType.BARE_METAL_SERVER.get_instance(
                client, instance_id
            )

        if not instance:
            raise IBMException(f"Instance not found: {instance_id}")

        return cls.from_existing(
            *args,
            client=client,
            instance=instance,
            **kwargs,
        )

    @staticmethod
    def create_raw_instance(
        client: VpcV1,
        *,
        name: str,
        image_id: str,
        vpc: VPC,
        instance_type: str,
        resource_group_id: str,
        zone: str,
        user_data=None,
        key_id: str,
    ) -> dict:
        """Create and return a raw IBM instance."""
        ibm_instance_type = _IBMInstanceType.from_instance_type(instance_type)

        initialization = {
            "image": {"id": image_id},
            "keys": [{"id": key_id}],
        }

        base_proto = {
            "name": name,
            "primary_network_interface": {
                "name": "eth0",
                "subnet": {"id": vpc.subnet_id},
            },
            "zone": {"name": zone},
            "profile": {"name": instance_type},
            "resource_group": {"id": resource_group_id},
            "vpc": {"id": vpc.id},
        }

        kwargs: dict
        if ibm_instance_type == _IBMInstanceType.VSI:
            instance_prototype: dict = {
                **base_proto,
                **initialization,
                "metadata_service": {"enabled": True},
            }
            if user_data:
                instance_prototype["user_data"] = user_data

            kwargs = {"instance_prototype": instance_prototype}

        elif ibm_instance_type == _IBMInstanceType.BARE_METAL_SERVER:
            kwargs = {
                **base_proto,
                "initialization": initialization,
            }

            if user_data:
                kwargs["initialization"]["user_data"] = user_data

        else:
            raise NotImplementedError(f"Implement me for: {ibm_instance_type}")

        raw_instance = ibm_instance_type.create_instance(
            client, **kwargs
        ).get_result()
        return raw_instance

    @staticmethod
    def _discover_floating_ip(client: VpcV1, instance: dict) -> Optional[dict]:
        """Discover a floating ip associated to instance."""
        nic_id = instance["primary_network_interface"]["id"]

        ibm_instance_type = _IBMInstanceType.from_raw_instance(instance)
        floating_ips = (
            ibm_instance_type.list_instance_network_interface_floating_ips(
                client,
                instance["id"],
                network_interface_id=nic_id,
            ).get_result()["floating_ips"]
        )

        return floating_ips[0] if floating_ips else None

    @property
    def name(self) -> str:
        """Return instance name."""
        return str(self._instance["name"])

    @property
    def ip(self):
        """Return IP address of instance."""
        if self._floating_ip is None:
            self._floating_ip = self._discover_floating_ip(
                self._client, self._refresh_instance()
            )
        if self._floating_ip is not None:
            return self._floating_ip["address"]

        self._log.warning(
            "The instance Instance %s has no IP available", self.id
        )
        return None

    @property
    def id(self) -> str:
        """Instance ID."""
        return str(self._instance["id"])

    @property
    def _floating_ip_id(self):
        return self._floating_ip["id"]

    @property
    def boot_volume_id(self) -> str:
        """Boot volume ID."""
        return self._instance["boot_volume_attachment"]["volume"]["id"]

    @property
    def _nic_id(self):
        return self._instance["primary_network_interface"]["id"]

    def delete(self, wait=True):
        """Delete the instance.

        Args:
            wait: wait for instance to be deleted
        """
        self._delete_instance(self.id)
        self._log.debug("deleting instance %s", self.id)
        if wait:
            self.wait_for_delete()

        self._client.delete_floating_ip(self._floating_ip_id)

    def _refresh_instance(self) -> dict:
        self._instance = self._get_instance(self.id).get_result()
        return self._instance

    def _wait_for_status(self, status: _Status, sleep_seconds: int = 300):
        _wait_until(
            lambda: self._refresh_instance()["status"] == status.value,
            timeout_seconds=sleep_seconds,
            timeout_msg_fn=lambda: (
                "Expected {status.value} state, but found"
                f" {self._instance['status']} "
                f"after waiting {sleep_seconds} seconds. "
                "Check IBM VPC console for more details."
            ),
        )

    def _do_restart(self, **kwargs):
        self._log.debug("restarting instance %s", self.id)
        self._execute_instance_action(action=_Action.REBOOT)

    def shutdown(self, wait=True, **kwargs):
        """Shutdown the instance.

        Args:
            wait: wait for the instance to shutdown
        """
        self._log.debug("shutting down instance %s", self.id)
        self._execute_instance_action(action=_Action.STOP)
        if wait:
            self.wait_for_stop()

    def start(self, wait=True):
        """Start the instance.

        Args:
            wait: wait for the instance to start.
        """
        self._execute_instance_action(action=_Action.START)
        if wait:
            self.wait()

    def _wait_for_instance_start(self):
        """Wait for the cloud instance to be up."""
        self._wait_for_status(_Status.RUNNING)

    def wait_for_delete(self, sleep_seconds=30, raise_on_fail=False):
        """Wait for instance to be deleted."""

        def _check_fn():
            try:
                self._refresh_instance()
            except ApiException as e:
                if e.code == 404:
                    return True  # Instance deleted
                raise
            return False

        msg = (
            f"Instance not terminated after {sleep_seconds} seconds. "
            "Check IBM VPC console."
        )

        terminated = _wait_until(
            _check_fn,
            timeout_seconds=sleep_seconds,
            timeout_msg_fn=lambda: msg,
            raise_on_fail=raise_on_fail,
        )
        if not terminated:
            self._log.warning(msg)

    def wait_for_stop(self):
        """Wait for instance stop."""
        self._wait_for_status(_Status.STOPPED)



"""Tests for pycloudlib.lxd.instance."""
import re
from copy import deepcopy
from json import dumps
from unittest import mock

import pytest

from pycloudlib.errors import PycloudlibTimeoutError
from pycloudlib.lxd.instance import LXDInstance, LXDVirtualMachineInstance
from pycloudlib.result import Result

LXD_QUERY = {
    "state": {
        "network": {
            "enp5s0": {
                "addresses": [
                    {
                        "address": "10.161.80.57",
                        "family": "inet",
                        "netmask": "24",
                        "scope": "global",
                    },
                    {
                        "address": "fd42:80e2:4695:1e96:216:3eff:fe06:e5f6",
                        "family": "inet6",
                        "netmask": "64",
                        "scope": "global",
                    },
                    {
                        "address": "fe80::216:3eff:fe06:e5f6",
                        "family": "inet6",
                        "netmask": "64",
                        "scope": "link",
                    },
                ],
                "counters": {
                    "bytes_received": 627023316,
                    "bytes_sent": 5159667,
                    "errors_received": 0,
                    "errors_sent": 0,
                    "packets_dropped_inbound": 0,
                    "packets_dropped_outbound": 0,
                    "packets_received": 344183,
                    "packets_sent": 71759,
                },
                "host_name": "tap55cb7af1",
                "hwaddr": "00:16:3e:06:e5:f6",
                "mtu": 1500,
                "state": "up",
                "type": "broadcast",
            },
            "lo": {
                "addresses": [
                    {
                        "address": "127.0.0.1",
                        "family": "inet",
                        "netmask": "8",
                        "scope": "local",
                    },
                    {
                        "address": "::1",
                        "family": "inet6",
                        "netmask": "128",
                        "scope": "local",
                    },
                ],
                "counters": {
                    "bytes_received": 67612,
                    "bytes_sent": 67612,
                    "errors_received": 0,
                    "errors_sent": 0,
                    "packets_dropped_inbound": 0,
                    "packets_dropped_outbound": 0,
                    "packets_received": 654,
                    "packets_sent": 654,
                },
                "host_name": "",
                "hwaddr": "",
                "mtu": 65536,
                "state": "up",
                "type": "loopback",
            },
            "veth1998ea41": {
                "addresses": [],
                "counters": {
                    "bytes_received": 100604,
                    "bytes_sent": 13587,
                    "errors_received": 0,
                    "errors_sent": 0,
                    "packets_dropped_inbound": 0,
                    "packets_dropped_outbound": 0,
                    "packets_received": 1210,
                    "packets_sent": 42,
                },
                "host_name": "",
                "hwaddr": "56:f1:b2:7f:8b:32",
                "mtu": 1500,
                "state": "up",
                "type": "broadcast",
            },
        },
        "pid": 182418,
        "processes": 46,
        "status": "Running",
        "status_code": 103,
    },
}

LXD_QUERY_BOND_BRIDGE = {
    "state": {
        "network": {
            "bond0": {
                "addresses": [],
                "counters": {
                    "bytes_received": 284039,
                    "bytes_sent": 32537,
                    "errors_received": 0,
                    "errors_sent": 0,
                    "packets_dropped_inbound": 0,
                    "packets_dropped_outbound": 0,
                    "packets_received": 419,
                    "packets_sent": 332,
                },
                "host_name": "tapd24998e9",
                "hwaddr": "02:00:00:56:ca:9f",
                "mtu": 1500,
                "state": "up",
                "type": "broadcast",
            },
            "enp5s0": {
                "addresses": [],
                "counters": {
                    "bytes_received": 284039,
                    "bytes_sent": 32537,
                    "errors_received": 0,
                    "errors_sent": 0,
                    "packets_dropped_inbound": 0,
                    "packets_dropped_outbound": 0,
                    "packets_received": 419,
                    "packets_sent": 332,
                },
                "host_name": "tapd24998e9",
                "hwaddr": "02:00:00:56:ca:9f",
                "mtu": 1500,
                "state": "up",
                "type": "broadcast",
            },
            "lo": {
                "addresses": [
                    {
                        "address": "127.0.0.1",
                        "family": "inet",
                        "netmask": "8",
                        "scope": "local",
                    },
                    {
                        "address": "::1",
                        "family": "inet6",
                        "netmask": "128",
                        "scope": "local",
                    },
                ],
                "counters": {
                    "bytes_received": 8280,
                    "bytes_sent": 8280,
                    "errors_received": 0,
                    "errors_sent": 0,
                    "packets_dropped_inbound": 0,
                    "packets_dropped_outbound": 0,
                    "packets_received": 104,
                    "packets_sent": 104,
                },
                "host_name": "",
                "hwaddr": "",
                "mtu": 65536,
                "state": "up",
                "type": "loopback",
            },
            "ovs-br": {
                "addresses": [
                    {
                        "address": "10.96.250.88",
                        "family": "inet",
                        "netmask": "24",
                        "scope": "global",
                    },
                    {
                        "address": "fd42:54d5:33be:7862:8da:baff:fe4e:dd4a",
                        "family": "inet6",
                        "netmask": "64",
                        "scope": "global",
                    },
                    {
                        "address": "fe80::ff:fe56:ca9f",
                        "family": "inet6",
                        "netmask": "64",
                        "scope": "link",
                    },
                ],
                "counters": {
                    "bytes_received": 279989,
                    "bytes_sent": 30289,
                    "errors_received": 0,
                    "errors_sent": 0,
                    "packets_dropped_inbound": 0,
                    "packets_dropped_outbound": 0,
                    "packets_received": 448,
                    "packets_sent": 304,
                },
                "host_name": "",
                "hwaddr": "0a:da:ba:4e:dd:4a",
                "mtu": 1500,
                "state": "up",
                "type": "broadcast",
            },
            "ovs-br.100": {
                "addresses": [
                    {
                        "address": "fe80::8da:baff:fe4e:dd4a",
                        "family": "inet6",
                        "netmask": "64",
                        "scope": "link",
                    }
                ],
                "counters": {
                    "bytes_received": 0,
                    "bytes_sent": 1076,
                    "errors_received": 0,
                    "errors_sent": 0,
                    "packets_dropped_inbound": 0,
                    "packets_dropped_outbound": 0,
                    "packets_received": 0,
                    "packets_sent": 14,
                },
                "host_name": "",
                "hwaddr": "0a:da:ba:4e:dd:4a",
                "mtu": 1500,
                "state": "up",
                "type": "broadcast",
            },
            "ovs-br.200": {
                "addresses": [
                    {
                        "address": "fe80::8da:baff:fe4e:dd4a",
                        "family": "inet6",
                        "netmask": "64",
                        "scope": "link",
                    }
                ],
                "counters": {
                    "bytes_received": 0,
                    "bytes_sent": 1146,
                    "errors_received": 0,
                    "errors_sent": 0,
                    "packets_dropped_inbound": 0,
                    "packets_dropped_outbound": 0,
                    "packets_received": 0,
                    "packets_sent": 15,
                },
                "host_name": "",
                "hwaddr": "0a:da:ba:4e:dd:4a",
                "mtu": 1500,
                "state": "up",
                "type": "broadcast",
            },
            "ovs-system": {
                "addresses": [],
                "counters": {
                    "bytes_received": 0,
                    "bytes_sent": 0,
                    "errors_received": 0,
                    "errors_sent": 0,
                    "packets_dropped_inbound": 0,
                    "packets_dropped_outbound": 0,
                    "packets_received": 0,
                    "packets_sent": 0,
                },
                "host_name": "",
                "hwaddr": "d6:45:c7:56:47:43",
                "mtu": 1500,
                "state": "down",
                "type": "broadcast",
            },
        }
    }
}


class TestRestart:
    """Tests covering pycloudlib.lxd.instance.Instance.restart."""

    @pytest.mark.parametrize("force", (False, True))
    @mock.patch("pycloudlib.lxd.instance.subp")
    def test_restart_calls_lxc_cmd_with_force_param(self, m_subp, force):
        """Honor force param on restart."""
        instance = LXDInstance(name="my_vm")
        instance._do_restart(force=force)  # pylint: disable=protected-access
        if force:
            assert "--force" in m_subp.call_args[0][0]
        else:
            assert "--force" not in m_subp.call_args[0][0]

    @mock.patch("pycloudlib.lxd.instance.LXDInstance.shutdown")
    @mock.patch("pycloudlib.lxd.instance.subp")
    def test_restart_does_not_shutdown(self, _m_subp, m_shutdown):
        """Don't shutdown (stop) instance on restart."""
        instance = LXDInstance(name="my_vm")
        instance._do_restart()  # pylint: disable=protected-access
        assert not m_shutdown.called


class TestExecute:
    """Tests covering pycloudlib.lxd.instance.Instance.execute."""

    def test_all_rcs_acceptable_when_using_exec(self):
        """Test that we invoke util.subp with rcs=None for exec calls.

        rcs=None means that we will get a Result object back for all return
        codes, rather than an exception for non-zero return codes.
        """
        instance = LXDInstance(None, execute_via_ssh=False)
        with mock.patch("pycloudlib.lxd.instance.subp") as m_subp:
            instance.execute("some_command")
        assert 1 == m_subp.call_count
        args, kwargs = m_subp.call_args
        assert "exec" in args[0]
        assert kwargs.get("rcs", mock.sentinel.not_none) is None


class TestVirtualMachineXenialAgentOperations:  # pylint: disable=W0212
    """Tests covering pycloudlib.lxd.instance.LXDVirtualMachineInstance."""

    # Key information we want in the logs when using non-ssh Xenial instances.
    _missing_agent_msg = "missing lxd-agent"

    @mock.patch("pycloudlib.lxd.instance.subp")
    def test_exec_with_run_command_on_xenial_machine(self, m_subp, caplog):
        """Test exec does not work with xenial vm."""
        instance = LXDVirtualMachineInstance(
            None, execute_via_ssh=False, series="xenial"
        )

        instance._run_command(["test"], None)
        assert self._missing_agent_msg in caplog.text
        assert m_subp.call_count == 1

    @mock.patch("pycloudlib.lxd.instance.subp")
    def test_file_pull_with_agent_on_xenial_machine(self, m_subp, caplog):
        """Test file pull does not work with xenial vm."""
        instance = LXDVirtualMachineInstance(
            None, execute_via_ssh=False, series="xenial"
        )

        instance.pull_file("/some/file", "/some/local/file")
        assert self._missing_agent_msg in caplog.text
        assert m_subp.call_count == 1

    @mock.patch("pycloudlib.lxd.instance.subp")
    def test_file_push_with_agent_on_xenial_machine(self, m_subp, caplog):
        """Test file push does not work with xenial vm."""
        instance = LXDVirtualMachineInstance(
            None, execute_via_ssh=False, series="xenial"
        )

        instance.push_file("/some/file", "/some/local/file")
        assert self._missing_agent_msg in caplog.text
        assert m_subp.call_count == 1
        expected_msg = (
            "Many Xenial images do not support `lxc file push` due to missing"
            " lxd-agent: you may see unavoidable failures.\n"
            "See https://github.com/canonical/pycloudlib/issues/132 for"
            " details."
        )
        assert expected_msg in caplog.messages
        assert m_subp.call_count == 1


class TestIP:
    """Tests covering pycloudlib.lxd.instance.Instance.ip."""

    @pytest.mark.parametrize(
        "stdouts,stderr,return_code,sleeps,expected",
        (
            (
                ["unparseable"],
                "",
                0,
                150,
                PycloudlibTimeoutError(
                    "Unable to determine IP address after 150 retries."
                    " exit:0 stdout: unparseable stderr: "
                ),
            ),
            (  # retry on non-zero exit code
                [dumps(LXD_QUERY)],
                "",
                1,
                150,
                PycloudlibTimeoutError(
                    "Unable to determine IP address after 150 retries."
                    " exit:1 stdout:"
                ),
            ),
            (  # empty values will retry indefinitely
                [""],
                "",
                0,
                150,
                PycloudlibTimeoutError(
                    "Unable to determine IP address after 150 retries."
                    " exit:0 stdout:  stderr: "
                ),
            ),
            (  # only retry until success
                ["unparseable", dumps(LXD_QUERY)],
                "",
                0,
                1,
                "10.161.80.57",
            ),
            ([dumps(LXD_QUERY)], "", 0, 0, "10.161.80.57"),
        ),
    )
    @mock.patch("pycloudlib.lxd.instance.time.sleep")
    @mock.patch("pycloudlib.lxd.instance.subp")
    def test_ip_parses_ipv4_output_from_lxc(
        self, m_subp, m_sleep, stdouts, stderr, return_code, sleeps, expected
    ):
        """IPv4 output matches specific vm name from `lxc list`.

        Errors are retried and result in PycloudlibTimeoutError on failure.
        """
        if len(stdouts) > 1:
            m_subp.side_effect = [
                Result(stdout=out, stderr=stderr, return_code=return_code)
                for out in stdouts
            ]
        else:
            m_subp.return_value = Result(
                stdout=stdouts[0], stderr=stderr, return_code=return_code
            )
        instance = LXDInstance(name="my_vm")
        lxc_mock = mock.call(
            ["lxc", "query", "/1.0/instances/my_vm?recursion=1"]
        )
        if isinstance(expected, Exception):
            with pytest.raises(type(expected), match=re.escape(str(expected))):
                instance.ip  # pylint: disable=pointless-statement
            assert [lxc_mock] * sleeps == m_subp.call_args_list
        else:
            assert expected == instance.ip
            assert [lxc_mock] * (1 + sleeps) == m_subp.call_args_list
        assert sleeps == m_sleep.call_count

    def test_parse_ip(self):
        """Verify ipv4 parser."""
        assert "10.161.80.57" == LXDInstance(name="my_vm").parse_ip(LXD_QUERY)
        local = deepcopy(LXD_QUERY)
        local.get("state", {}).get("network", {}).pop("enp5s0")
        assert LXDInstance(name="my_vm").parse_ip(local) is None

    def test_parse_ip_bond_bridge(self):
        """Verify ipv4 parser with a cfg with bonds and bridging."""
        assert "10.96.250.88" == LXDInstance(name="my_vm").parse_ip(
            LXD_QUERY_BOND_BRIDGE
        )


class TestWaitForStop:
    """Tests covering pycloudlib.lxd.instance.Instance.wait_for_stop."""

    @pytest.mark.parametrize("is_ephemeral", ((True), (False)))
    def test_wait_for_stop_does_not_wait_for_ephemeral_instances(
        self, is_ephemeral
    ):
        """LXDInstance.wait_for_stop does not wait on ephemeral instances."""
        instance = LXDInstance(name="test")
        with mock.patch.object(instance, "wait_for_state") as wait_for_state:
            with mock.patch.object(type(instance), "ephemeral", is_ephemeral):
                instance.wait_for_stop()

        call_count = 0 if is_ephemeral else 1
        assert call_count == wait_for_state.call_count


class TestShutdown:
    """Tests covering pycloudlib.lxd.instance.Instance.shutdown."""

    @pytest.mark.parametrize(
        "wait,force,cmd",
        (
            (True, False, ["lxc", "stop", "test"]),
            (False, False, ["lxc", "stop", "test"]),
            (True, True, ["lxc", "stop", "test", "--force"]),
        ),
    )
    @mock.patch("pycloudlib.lxd.instance.subp")
    def test_shutdown_calls_wait_for_stopped_state_when_wait_true(
        self, m_subp, wait, force, cmd
    ):
        """LXDInstance.wait_for_stopped called when wait is True."""
        instance = LXDInstance(name="test")
        with mock.patch.object(instance, "wait_for_stop") as wait_for_stop:
            with mock.patch.object(type(instance), "state", "RUNNING"):
                instance.shutdown(wait=wait, force=force)

        assert [mock.call(cmd)] == m_subp.call_args_list
        call_count = 1 if wait else 0
        assert call_count == wait_for_stop.call_count


class TestDelete:
    """Tests covering pycloudlib.lxd.instance.Instance.delete."""

    @pytest.mark.parametrize("is_ephemeral", ((True), (False)))
    @mock.patch("pycloudlib.lxd.instance.LXDInstance.shutdown")
    @mock.patch("pycloudlib.lxd.instance.subp")
    def test_delete_on_ephemeral_instance_calls_shutdown(
        self, m_subp, m_shutdown, is_ephemeral
    ):
        """Check if ephemeral instance delete stops it instead of deleting it.

        Also verify is delete is actually called if instance is not ephemeral.
        """
        instance = LXDInstance(name="test")

        with mock.patch.object(type(instance), "ephemeral", is_ephemeral):
            instance.delete(wait=False)

        assert 0 == m_shutdown.call_count
        assert 1 == m_subp.call_count
        assert [
            mock.call(["lxc", "delete", "test", "--force"])
        ] == m_subp.call_args_list



# This file is part of pycloudlib. See LICENSE file for license information.
"""Used to define custom Virtual Private Clouds (VPC)."""

import ipaddress
import logging

from botocore.exceptions import ClientError

from pycloudlib.ec2.util import _tag_resource
from pycloudlib.errors import CloudError

logger = logging.getLogger(__name__)


class VPC:
    """Virtual Private Cloud class proxy for AWS VPC resource."""

    def __init__(self, vpc):
        """Create a VPC proxy instance for an AWS VPC resource.

        Args:
            vpc_id: Optional ID of existing VPC object to return
        """
        self.vpc = vpc

    @classmethod
    def create(cls, resource, name, ipv4_cidr="192.168.1.0/20"):
        """Create a pycloudlib.ec2.VPC proxy for an AWS VPC resource.

        Args:
            resource: EC2 resource client
            name: String for the name or tag of the VPC
            ipv4_cidr: String of the CIDR for IPV4 subnet to associate with the
                VPC.

        Returns:
            pycloudlib.ec2.VPC instance

        """
        logger.debug("Creating VPC named (%s)", name)
        vpc = cls._create_vpc(
            resource=resource, name=name, ipv4_cidr=ipv4_cidr
        )
        vpc.wait_until_available()
        vpc.reload()
        gateway = cls._create_internet_gateway(resource, vpc)
        subnet = cls._create_subnet(vpc, ipv4_cidr)
        route_table = cls._create_routing_table(vpc, gateway.id, subnet.id)
        sec_group = cls._create_security_group(vpc, name)
        for aws_resource in (vpc, gateway, subnet, route_table, sec_group):
            _tag_resource(aws_resource, name)
        logger.debug("Created VPC (%s) named (%s)", vpc.id, name)
        return cls(vpc)

    @classmethod
    def from_existing(cls, resource, vpc_id):
        """Wrap an existing boto3 EC2 VPC resource given the vpc_id.

        Args:
            resource: EC2 resource client
            vpc_id: String for an existing VPC id.

        Returns:
            pycloudlib.ec2.VPC instance

        """
        logger.debug("Reusing existing VPC (%s)", vpc_id)
        vpc = resource.Vpc(vpc_id)
        return cls(vpc)

    @property
    def id(self):
        """ID of the VPC."""
        return self.vpc.id

    @property
    def name(self):
        """Name of the VPC from tags."""
        for tag in self.vpc.tags:
            if tag["Key"] == "Name":
                return tag["Value"]
        return "NO-TAG-NAME-PRESENT"

    @classmethod
    def _create_internet_gateway(cls, resource, vpc):
        """Create Internet Gateway and assign to VPC.

        Returns:
            Internet gateway object

        """
        logger.debug("creating internet gateway for vpc %s", vpc.id)
        internet_gateway = resource.create_internet_gateway()
        internet_gateway.attach_to_vpc(VpcId=vpc.id)

        return internet_gateway

    @classmethod
    def _create_routing_table(cls, vpc, gateway_id, subnet_id):
        """Update default routing table with internet gateway and subnet.

        This sets up internet access between the VPC via the internet gateway
        by configuring routing tables for IPv4 and IPv6.
        """
        logger.debug("creating routing table")
        route_table = vpc.create_route_table()
        route_table.create_route(
            DestinationCidrBlock="0.0.0.0/0", GatewayId=gateway_id
        )
        route_table.create_route(
            DestinationIpv6CidrBlock="::/0", GatewayId=gateway_id
        )
        route_table.associate_with_subnet(SubnetId=subnet_id)
        return route_table

    @classmethod
    def _create_security_group(cls, vpc, name):
        """Enable ingress to default VPC security group.

        Returns:
            Security group object

        """
        logger.debug("creating security group")
        security_group = vpc.create_security_group(
            GroupName=name, Description="pycloudlib created security group"
        )
        security_group.authorize_ingress(
            IpProtocol="-1", FromPort=-1, ToPort=-1, CidrIp="0.0.0.0/0"
        )

        return security_group

    @classmethod
    def _create_subnet(cls, vpc, ipv4_cidr):
        """Generate IPv4 and IPv6 subnets for use in an AWS VPC resource.

        Args:
            vpc: AWS VPC resource to which the created subnet is associated.
            ipv4_cidr: CIDR for IPV4 network

        Returns:
            Create subnet object

        """
        ipv6_cidr = vpc.ipv6_cidr_block_association_set[0]["Ipv6CidrBlock"]
        kwargs = {"CidrBlock": ipv4_cidr}
        try:
            ipaddress.IPv6Network(ipv6_cidr)
            kwargs["Ipv6CidrBlock"] = ipv6_cidr[:-2] + "64"
        except ValueError as e:
            logger.warning(
                "Skipping IPv6 association on vpc."
                " Could not understand Ipv6CidrBlock: [%s]: %s",
                ipv6_cidr,
                str(e),
            )
        logger.debug("creating subnets with following ranges:")
        for key, value in kwargs.items():
            logger.debug("%s: %s", key, value)
        subnet = vpc.create_subnet(**kwargs)

        # enable public IP on instance launch
        modify_subnet = subnet.meta.client.modify_subnet_attribute
        modify_subnet(SubnetId=subnet.id, MapPublicIpOnLaunch={"Value": True})

        return subnet

    @classmethod
    def _create_vpc(cls, resource, name, ipv4_cidr):
        """Set up AWS EC2 VPC or return existing VPC.

        Args:
            resource: boto 3 resource client
            name: the name/tag of the VPC to create
            ipv4_cidr: CIDR for IPV4 network

        Returns:
            VPC resource created from AWS cli

        """
        logger.debug(
            "creating new vpc named %s with subnet %s", name, ipv4_cidr
        )
        try:
            vpc = resource.create_vpc(
                CidrBlock=ipv4_cidr, AmazonProvidedIpv6CidrBlock=True
            )
        except ClientError as error:
            raise CloudError(error) from error

        vpc.wait_until_available()

        return vpc

    def delete(self):
        """Terminate all associated instances and delete an entire VPC."""
        for instance in self.vpc.instances.all():
            logger.debug("waiting for instance %s termination", instance.id)
            instance.terminate()
            instance.wait_until_terminated()

        for security_group in self.vpc.security_groups.all():
            logger.debug("deleting security group %s", security_group.id)
            security_group.delete()

        for subnet in self.vpc.subnets.all():
            logger.debug("deleting subnet %s", subnet.id)
            subnet.delete()

        for route_table in self.vpc.route_tables.all():
            logger.debug("deleting routing table %s", route_table.id)
            route_table.delete()

        for gateway in self.vpc.internet_gateways.all():
            logger.debug("deleting internet gateway %s", gateway.id)
            gateway.detach_from_vpc(VpcId=self.vpc.id)
            gateway.delete()

        if self.vpc:
            logger.debug("deleting vpc %s", self.vpc.id)
            self.vpc.delete()



# This file is part of pycloudlib. See LICENSE file for license information.
"""EC2 instance."""
import string
import time

import botocore

from pycloudlib.errors import PycloudlibError
from pycloudlib.instance import BaseInstance


class EC2Instance(BaseInstance):
    """EC2 backed instance."""

    _type = "ec2"

    def __init__(self, key_pair, client, instance):
        """Set up instance.

        Args:
            key_pair: SSH key object
            client: boto3 client object
            instance: created boto3 instance object
        """
        super().__init__(key_pair)

        self._instance = instance
        self._ip = None
        self._client = client

        self.boot_timeout = 300

    def __repr__(self):
        """Create string representation for class."""
        return "{}(key_pair={}, client={}, instance={})".format(
            self.__class__.__name__,
            self.key_pair,
            self._client,
            self._instance,
        )

    @property
    def availability_zone(self):
        """Return availability zone."""
        return self._instance.placement["AvailabilityZone"]

    @property
    def ip(self):
        """Return IP address of instance."""
        self._instance.reload()
        return self._instance.public_ip_address

    @property
    def id(self):
        """Return id of instance."""
        return self._instance.instance_id

    @property
    def name(self):
        """Return id of instance."""
        return self.id

    @property
    def image_id(self):
        """Return id of instance."""
        return self._instance.image_id

    def add_network_interface(self) -> str:
        """Add network interface to instance.

        Creates an ENI device and attaches it to the running instance. This
        is effectively a hot-add of a network device. Returns the IP address
        of the added network interface as a string.

        See the AWS documentation for more info:
        https://boto3.readthedocs.io/en/latest/reference/services/ec2.html?#EC2.Client.create_network_interface
        https://boto3.readthedocs.io/en/latest/reference/services/ec2.html?#EC2.Client.attach_network_interface
        """
        self._log.debug("adding network interface to %s", self.id)
        interface_id = self._create_network_interface()
        return self._attach_network_interface(interface_id)

    def add_volume(self, size=8, drive_type="gp2"):
        """Add storage volume to instance.

        Creates an EBS volume and attaches it to the running instance. This
        is effectively a hot-add of a storage device.

        See AWS documentation for more info:
        https://boto3.readthedocs.io/en/latest/reference/services/ec2.html?#EC2.Client.create_volume
        https://boto3.readthedocs.io/en/latest/reference/services/ec2.html?#EC2.Client.attach_volume

        Args:
            size: Size in GB of the drive to add
            drive_type: Type of EBS volume to add
        """
        self._log.debug("adding storage volume to %s", self.id)
        volume = self._create_ebs_volume(size, drive_type)
        self._attach_ebs_volume(volume)

    def console_log(self):
        """Collect console log from instance.

        The console log is buffered and not always present, therefore
        may return empty string.

        Returns:
            The console log or error message

        """
        start = time.time()
        while time.time() < start + 300:
            response = self._instance.console_output(Latest=True)
            try:
                return response["Output"]
            except KeyError:
                self._log.debug("Console output not yet available; sleeping")
                time.sleep(5)
        return "No Console Output [%s]" % self._instance

    def delete(self, wait=True):
        """Delete instance."""
        self._log.debug("deleting instance %s", self._instance.id)
        self._instance.terminate()

        if wait:
            self.wait_for_delete()

    def _do_restart(self, **kwargs):
        """Restart the instance."""
        self._log.debug("restarting instance %s", self._instance.id)
        self._instance.reboot()

    def shutdown(self, wait=True, **kwargs):
        """Shutdown the instance.

        Args:
            wait: wait for the instance shutdown
        """
        self._log.debug("shutting down instance %s", self._instance.id)
        self._instance.stop()

        if wait:
            self.wait_for_stop()

    def start(self, wait=True):
        """Start the instance.

        Args:
            wait: wait for the instance to start.
        """
        if self._instance.state["Name"] == "running":
            return

        self._log.debug("starting instance %s", self._instance.id)
        self._instance.start()

        if wait:
            self.wait()

    def _wait_for_instance_start(self):
        """Wait for instance to be up."""
        self._log.debug("wait for instance running %s", self._instance.id)
        self._instance.wait_until_running()
        self._log.debug("reloading instance state %s", self._instance.id)
        self._instance.reload()

    def wait_for_delete(self):
        """Wait for instance to be deleted."""
        self._instance.wait_until_terminated()
        self._instance.reload()

    def wait_for_stop(self):
        """Wait for instance stop."""
        self._instance.wait_until_stopped()
        self._instance.reload()

    def _attach_ebs_volume(self, volume):
        """Attach EBS volume to an instance.

        The volume will get added at the next available volume name.

        The volume will also be set to delete on termination of the
        instance.

        Args:
            volume: boto3 volume object
        """
        mount_point = self._get_free_volume_name()
        args = {
            "Device": mount_point,
            "InstanceId": self.id,
            "VolumeId": volume["VolumeId"],
        }

        self._client.attach_volume(**args)

        waiter = self._client.get_waiter("volume_in_use")
        waiter.wait(VolumeIds=[volume["VolumeId"]])

        self._instance.reload()

        self._instance.modify_attribute(
            BlockDeviceMappings=[
                {
                    "DeviceName": mount_point,
                    "Ebs": {"DeleteOnTermination": True},
                }
            ]
        )

    def _attach_network_interface(self, interface_id: str) -> str:
        """Attach ENI device to an instance.

        This will attach the interface at the next available index.

        The device will also be set to delete on termination of the
        instance.

        Args:
            interface_id: string, id of interface to attach
        Returns:
            IP address of the added interface
        """
        device_index = self._get_free_nic_index()
        args = {
            "DeviceIndex": device_index,
            "InstanceId": self.id,
            "NetworkInterfaceId": interface_id,
        }

        response = self._client.attach_network_interface(**args)

        self._instance.reload()

        for nic in self._instance.network_interfaces:
            if nic.attachment["AttachmentId"] == response["AttachmentId"]:
                nic.modify_attribute(
                    Attachment={
                        "AttachmentId": response["AttachmentId"],
                        "DeleteOnTermination": True,
                    }
                )
                return nic.private_ip_address
        raise PycloudlibError(
            "Could not attach NIC with AttachmentId: {}".format(
                response.get("AttachmentId", None)
            )
        )

    def _create_ebs_volume(self, size, drive_type):
        """Create EBS volume.

        Args:
            size: Size of drive to create in GB
            drive_type: Type of drive to create

        Returns:
            The boto3 volume object

        """
        args = {
            "AvailabilityZone": self.availability_zone,
            "Size": size,
            "VolumeType": drive_type,
            "TagSpecifications": [
                {
                    "ResourceType": "volume",
                    "Tags": [{"Key": "Name", "Value": self.id}],
                }
            ],
        }

        volume = self._client.create_volume(**args)

        waiter = self._client.get_waiter("volume_available")
        waiter.wait(VolumeIds=[volume["VolumeId"]])

        return volume

    def _create_network_interface(self) -> str:
        """Create ENI device.

        Returns:
            The ENI device id

        """
        args = {
            "Groups": [
                group["GroupId"] for group in self._instance.security_groups
            ],
            "SubnetId": self._instance.subnet_id,
        }

        response = self._client.create_network_interface(**args)
        interface_id = response["NetworkInterface"]["NetworkInterfaceId"]

        waiter = self._client.get_waiter("network_interface_available")
        waiter.wait(NetworkInterfaceIds=[interface_id])

        return interface_id

    def _get_free_nic_index(self) -> int:
        """Determine a free NIC interface for an instance.

        Per the following doc the maximum number of NICs is 16:
        https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/using-eni.html

        Returns:
            integer to use as index for NIC

        """
        used_indexes = [
            nic.attachment["DeviceIndex"]
            for nic in self._instance.network_interfaces
        ]
        for possible_index in range(16):
            if possible_index not in used_indexes:
                return possible_index
        raise PycloudlibError("No free nics left!")

    def _get_free_volume_name(self):
        """Determine a free volume mount point for an instance.

        Loop through used mount names (e.g. /dev/sda1, /dev/sdb) and
        the possible device names (e.g. /dev/sdf, /dev/sdg... /dev/sdz)
        and find the first that is available.

        This also works for instances which only have NVMe devices or
        when mounting NVMe EBS volumes. In which case, this suggestion
        is ignored an the number number is used.

        Using /dev/sd* per the following doc:
        https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/device_naming.html

        Returns:
            string of latest name available

        """
        all_device_names = []
        for name in string.ascii_lowercase:
            if name not in "abcde":
                all_device_names.append("/dev/sd%s" % name)

        used_device_names = set()
        for device in self._instance.block_device_mappings:
            used_device_names.add(device["DeviceName"])

        return list(set(all_device_names) - used_device_names)[0]

    def remove_network_interface(self, ip_address):
        """Remove network interface based on IP address.

        Find the NIC from the IP, detach from the instance, then delete the
        NIC.
        """
        # Get the NIC from the IP
        nic = [
            nic
            for nic in self._instance.network_interfaces
            if nic.private_ip_address == ip_address
        ][0]
        self._client.detach_network_interface(
            AttachmentId=nic.attachment["AttachmentId"]
        )

        # Detach from the instance
        for _ in range(60):
            self._instance.reload()
            nics = [
                nic
                for nic in self._instance.network_interfaces
                if nic.id == ip_address
            ]
            if not nics:
                break
            time.sleep(1)
        else:
            raise PycloudlibError("Network interface did not detach")

        # Delete the NIC
        try:
            self._client.delete_network_interface(NetworkInterfaceId=nic.id)
        except botocore.exceptions.ClientError:
            self._log.debug(
                "Failed manually deleting network interface. "
                "Interface should get destroyed on instance cleanup."
            )
